__author__ = 'labx'

import sys, os, numpy, platform
import orangecanvas.resources as resources
from PyQt5 import QtGui, QtCore

try:
    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from matplotlib import figure as matfig
    import pylab
except ImportError:
    print(sys.exc_info()[1])
    pass


import xraylib

from oasys.widgets import gui

class EmittingStream(QtCore.QObject):
    textWritten = QtCore.pyqtSignal(str)

    def write(self, text):
        self.textWritten.emit(str(text))

class XoppyPlot:

    @classmethod
    def plot_histo(cls, plot_window, x, y, title, xtitle, ytitle, color='blue', replace=True):
        matplotlib.rcParams['axes.formatter.useoffset']='False'
        plot_window.addCurve(x, y, title, symbol='', color=color, xlabel=xtitle, ylabel=ytitle, replace=replace) #'+', '^', ','
        if not xtitle is None: plot_window.setGraphXLabel(xtitle)
        if not ytitle is None: plot_window.setGraphYLabel(ytitle)
        if not title is None: plot_window.setGraphTitle(title)

        # plot_window.setDrawModeEnabled(True, 'rectangle')
        plot_window.setInteractiveMode('zoom',color='orange')
        plot_window.resetZoom()
        plot_window.replot()

        plot_window.setActiveCurve(title)



import os, sys
import numpy

from PyQt5.QtCore import QRect, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox, QScrollArea, QTableWidget, QTableWidgetItem, QHeaderView, QAbstractItemView, QWidget, QLabel, QSizePolicy
from PyQt5.QtGui import QTextCursor,QFont, QPalette, QColor, QPainter, QBrush, QPen, QPixmap
from PyQt5.QtWidgets import QLabel, QSizePolicy
from PyQt5.QtWidgets import QApplication
from PyQt5 import QtGui

from silx.gui.plot import Plot2D

import orangecanvas.resources as resources

from orangewidget import gui, widget
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence

from oasys.util.oasys_objects import OasysSurfaceData
from oasys.util.oasys_util import write_surface_file
from oasys.util.oasys_util import EmittingStream

from shadow4.optical_surfaces.s4_conic import S4Conic

from orangecontrib.esrf.shadow.util.conics_from_factory_parameters import ellipsoid, hyperboloid, \
    paraboloid_focusing, paraboloid_collimating

from orangecontrib.esrf.shadow.util.conics_from_factory_parameters import ken_ellipsoid, ken_hyperboloid, \
    ken_paraboloid_focusing, ken_paraboloid_collimating

from orangecontrib.esrf.shadow.util.conic_penelope import ellipsoid as p_ellipsoid
from orangecontrib.esrf.shadow.util.conic_penelope import paraboloid as p_paraboloid
from orangecontrib.esrf.shadow.util.conic_penelope import sphere as p_sphere
from orangecontrib.esrf.shadow.util.conic_penelope import hyperboloid as p_hyperboloid

# from orangecontrib.syned.util.diaboloid_tools import diaboloid_approximated_point_to_segment
# from orangecontrib.syned.util.diaboloid_tools import diaboloid_approximated_segment_to_point
# from orangecontrib.syned.util.diaboloid_tools import diaboloid_exact_point_to_segment
# from orangecontrib.syned.util.diaboloid_tools import diaboloid_exact_segment_to_point
# from orangecontrib.syned.util.diaboloid_tools import parabolic_cone_point_to_segment
# from orangecontrib.syned.util.diaboloid_tools import parabolic_cone_segment_to_point
# from orangecontrib.syned.util.diaboloid_tools import parabolic_cone_linearized_point_to_segment
# from orangecontrib.syned.util.diaboloid_tools import parabolic_cone_linearized_segment_to_point
# from orangecontrib.syned.util.diaboloid_tools import toroid_point_to_segment, toroid_segment_to_point

class OWConic(OWWidget):
    name = "Conic surface"
    id = "conic"
    description = "Conic surface generator"
    icon = "icons/conic.png"
    author = "M Sanchez del Rio"
    maintainer_email = "srio@esrf.eu"
    priority = 15
    category = ""
    keywords = ["preprocessor", "surface", "conic"]

    outputs = [{"name": "Surface Data",
                "type": OasysSurfaceData,
                "doc": "Surface Data",
                "id": "Surface Data"},
               ]

    want_main_area = 1
    want_control_area = 1

    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 645

    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 650

    #
    # variable list
    #

    configuration = Setting(5)
    calculation_method = Setting(0)
    source_oe = Setting(3.0)
    oe_image = Setting(10.0)
    theta = Setting(0.003) # mrad
    ny = Setting(1001)
    nx = Setting(101)
    semilength_x = Setting(0.015)
    semilength_y = Setting(0.25)
    # detrend_toroid = Setting(0)
    filename_h5 = Setting("conic.h5")

    cylindrize = Setting(0)

    #
    #
    #

    tab=[]
    usage_path = os.path.join(resources.package_dirname("orangecontrib.esrf.syned.widgets.extension") , "misc", "conic_usage.png")

    def __init__(self):
        super().__init__()

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width() * 0.05),
                               round(geom.height() * 0.05),
                               round(min(geom.width() * 0.98, self.MAX_WIDTH)),
                               round(min(geom.height() * 0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        gui.separator(self.controlArea)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)


        tab_calc = oasysgui.createTabPage(tabs_setting, "Calculate")
        tab_usa = oasysgui.createTabPage(tabs_setting, "Use of the Widget")

        #
        #-------------------- calculate
        #

        button = gui.button(tab_calc, self, "Calculate", callback=self.calculate)

        out_calc = oasysgui.widgetBox(tab_calc, "Conic Surface Parameters", addSpace=True, orientation="vertical")

        gui.comboBox(out_calc, self, "configuration", label="Focusing configuration", labelWidth=300,
                     items=self.get_surface_list(),
                     sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(out_calc, self, "calculation_method", label="calculation method", labelWidth=300,
                     items=["s4_conic", "penelope", "mathematica", "Ken"], sendSelectedValue=False, orientation="horizontal")


        oasysgui.lineEdit(out_calc, self, "source_oe", "distance source to mirror [m]",
                           labelWidth=300, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(out_calc, self, "oe_image", "distance mirror to image [m]",
                           labelWidth=300, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(out_calc, self, "theta", "grazing angle [rad]",
                           labelWidth=300, valueType=float, orientation="horizontal")


        #
        # --------------- MESH
        #
        out_calc = oasysgui.widgetBox(tab_calc, "Mesh Parameters", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(out_calc, self, "ny", "Points in Y (tangential)",
                           labelWidth=300, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(out_calc, self, "nx", "Points in X (sagittal)",
                           labelWidth=300, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(out_calc, self, "semilength_y", "Half length Y [m]",
                           labelWidth=300, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(out_calc, self, "semilength_x", "Half length X [m]",
                           labelWidth=300, valueType=float, orientation="horizontal")

        #
        # --------------- MODIFICATIONS
        #
        out_calc = oasysgui.widgetBox(tab_calc, "Modify surface", addSpace=True, orientation="vertical")
        gui.comboBox(out_calc, self, "cylindrize", label="Cylindrize", labelWidth=300,
                     items=["No [default]", "Yes [meridional]", "Yes [sagittal]"], sendSelectedValue=False,
                     orientation="horizontal")

        gui.separator(out_calc)

        #
        # --------------- FILE
        #
        out_file = oasysgui.widgetBox(tab_calc, "Output hdf5 file", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(out_file , self, "filename_h5", "Output filename *.h5",
                           labelWidth=150, valueType=str, orientation="horizontal")

        gui.separator(out_file)

        #
        #-------------------- Use
        #

        tab_usa.setStyleSheet("background-color: white;")

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)

        #
        #
        #
        gui.rubber(self.controlArea)
        self.initializeTabs()
        gui.rubber(self.mainArea)

    def get_surface_list(self):
        return ["Plane",
                "Paraboloid (collimating)",
                "Paraboloid (focusing)",
                "Sphere",
                "Ellipsoid",
                "Hyperboloid",
                ]
    def initializeTabs(self):
        self.tabs = oasysgui.tabWidget(self.mainArea)

        self.tab = [oasysgui.createTabPage(self.tabs, "Results"),
                    oasysgui.createTabPage(self.tabs, "Output"),
        ]

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.plot_canvas = [None] * len(self.tab)

        # tab index 1
        self.profileInfo = oasysgui.textArea()
        profile_box = oasysgui.widgetBox(self.tab[1], "", addSpace=True, orientation="horizontal")
        profile_box.layout().addWidget(self.profileInfo)

        for index in range(len(self.tab)):
            try:
                self.tab[index].layout().addWidget(self.plot_canvas[index])
            except:
                pass
        self.tabs.setCurrentIndex(0)

    def check_fields(self):
        self.nx = congruence.checkStrictlyPositiveNumber(self.nx, "Points X")
        self.ny = congruence.checkStrictlyPositiveNumber(self.ny, "Points Y")
        self.theta = congruence.checkStrictlyPositiveNumber(self.theta, "Grazing angle")
        self.semilength_x = congruence.checkStrictlyPositiveNumber(self.semilength_x, "Half length X")
        self.semilength_y = congruence.checkStrictlyPositiveNumber(self.semilength_y, "Half length Y")
        self.source_ow = congruence.checkNumber(self.source_oe, "Distance source-mirror")
        self.oe_image = congruence.checkNumber(self.oe_image, "Distance mirror-image")

    def writeStdOut(self, text="", initialize=False):
        cursor = self.profileInfo.textCursor()
        if initialize:
            self.profileInfo.setText(text)
        else:
            cursor.movePosition(QtGui.QTextCursor.End)
            cursor.insertText(text)

    def calculate(self):
        self.writeStdOut(initialize=True)
        sys.stdout = EmittingStream(textWritten=self.writeStdOut)

        self.check_fields()

        x = numpy.linspace(-self.semilength_x, self.semilength_x, self.nx)
        y = numpy.linspace(-self.semilength_y, self.semilength_y, self.ny)

        X = numpy.outer(x, numpy.ones_like(y))
        Y = numpy.outer(numpy.ones_like(x), y)

        p = self.source_oe
        q = self.oe_image
        theta = self.theta
        print("Inputs: p=%g m, q=%g m, theta=%g rad: " % (p, q, theta))

        # "Plane",
        # "Paraboloid (collimating)",
        # "Paraboloid (focusing)",
        # "Sphere",
        # "Ellipsoid",
        # "Hyperboloid",

        mirror_txt = "Conic"
        if self.get_surface_list()[self.configuration] == "Plane": #
            print("Method=s4_conic(**fixed**)")
            s4 = S4Conic.initialize_as_plane()
        elif self.get_surface_list()[self.configuration] == "Paraboloid (collimating)":  #
            if self.calculation_method == 0:
                print("Method=s4_conic")
                s4 = S4Conic.initialize_as_paraboloid_from_focal_distances(self.source_oe, 1e10, self.theta,
                                                            cylindrical=0, cylangle=0.0, switch_convexity=0)
            elif self.calculation_method == 1:
                print("Method=penelope")
                ccc = p_paraboloid(ssour=self.source_oe, simag=1e10, theta_grazing=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc['ccc'])
            elif self.calculation_method == 2:
                print("Method=mathematica")
                ccc = paraboloid_collimating(p=self.source_oe, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
            elif self.calculation_method == 3:
                print("Method=ken")
                ccc = ken_paraboloid_collimating(p=self.source_oe, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
        elif self.get_surface_list()[self.configuration] == "Paraboloid (focusing)":  #
            if self.calculation_method == 0:
                print("Method=s4_conic")
                s4 = S4Conic.initialize_as_paraboloid_from_focal_distances(1e10, self.oe_image, self.theta,
                                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
            elif self.calculation_method == 1:
                print("Method=penelope")
                ccc = p_paraboloid(ssour=1e10, simag=self.oe_image, theta_grazing=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc['ccc'])
            elif self.calculation_method == 2:
                print("Method=mathematica")
                ccc = paraboloid_focusing(q=self.oe_image, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
            elif self.calculation_method == 3:
                print("Method=ken")
                ccc = ken_paraboloid_focusing(q=self.oe_image, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
        elif self.get_surface_list()[self.configuration] == "Sphere":  #
            if self.calculation_method == 1:
                print("Method=penelope")
                ccc = p_sphere(ssour=self.source_oe, simag=self.oe_image, theta_grazing=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc['ccc'])
            else:
                print("Method=s4_conic(**fixed**)")
                s4 = S4Conic.initialize_as_sphere_from_focal_distances(self.source_oe, self.oe_image, self.theta,
                                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
        elif self.get_surface_list()[self.configuration] == "Ellipsoid":  #
            if self.calculation_method == 0:
                print("Method=s4_conic")
                s4 = S4Conic.initialize_as_ellipsoid_from_focal_distances(self.source_oe, self.oe_image, self.theta,
                                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
            elif self.calculation_method == 1:
                print("Method=penelope")
                ccc = p_ellipsoid(ssour=self.source_oe, simag=self.oe_image, theta_grazing=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc['ccc'])
            elif self.calculation_method == 2:
                print("Method=mathematica")
                ccc = ellipsoid(p=self.source_oe, q=self.oe_image, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
            elif self.calculation_method == 3:
                print("Method=ken")
                ccc = ken_ellipsoid(p=self.source_oe, q=self.oe_image, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
        elif self.get_surface_list()[self.configuration] == "Hyperboloid":  #
            if self.calculation_method == 0:
                print("Method=s4_conic")
                s4 = S4Conic.initialize_as_hyperboloid_from_focal_distances(self.source_oe, self.oe_image, self.theta,
                                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
            elif self.calculation_method == 1:
                print("Method=penelope")
                ccc = p_hyperboloid(ssour=self.source_oe, simag=self.oe_image, theta_grazing=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc['ccc'])
            elif self.calculation_method == 2:
                print("Method=mathematica")
                ccc = hyperboloid(p=self.source_oe, q=self.oe_image, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
            elif self.calculation_method == 3:
                print("Method=ken")
                ccc = ken_hyperboloid(p=self.source_oe, q=self.oe_image, theta=self.theta)
                s4 = S4Conic.initialize_from_coefficients(ccc)
        else:
            raise Exception("Not implemented")

        #
        # cyl?
        #
        if self.cylindrize == 0:
            pass
        elif self.cylindrize == 1:
            s4.set_cylindrical(0)
        elif self.cylindrize == 2:
            s4.set_cylindrical(numpy.pi/2)

        #
        # display coeffs
        #
        ccc = s4.get_coefficients()
        print("\n\n\nConic coefficients: ")
        for i in range(10):
            print("    ccc[%d] = %20.12g" % (i, ccc[i]))

        print("\nConic coefficients (normalized): ")
        for i in range(10):
            print("    ccc[%d] = %20.12g" % (i, ccc[i]/ccc[0]))

        #
        # numerical surface
        #

        Z = s4.height(y=Y, x=X, return_solution=0)

        write_surface_file(Z.T, x, y, self.filename_h5, overwrite=True)
        print("\nHDF5 file %s written to disk." % self.filename_h5)



        # Z -= Ztor
        #

        #
        #
        # if self.detrend_toroid == 0:
        #     Ztor = 0
        # elif self.detrend_toroid == 1:  # detrend toroid
        #     mirror_txt += " (sphere removed)"
        #     Ztor = 0
        # else:
        #     raise Exception("Not implemented")


        self.plot_data2D(Z, x, y, self.tab[0],
                         title="%s p:%6.3f m, q:%6.3f %6.3f mrad" %
                               (mirror_txt, self.source_oe, self.oe_image, self.theta),
                         xtitle="x (sagittal) [m] (%d pixels)" % x.size,
                         ytitle="y (tangential) [m] (%d pixels)" % y.size)




        self.send("Surface Data",
                  OasysSurfaceData(xx=x,
                                   yy=y,
                                   zz=Z.T,
                                   surface_data_file=self.filename_h5))


    def plot_data2D(self, data2D, dataX, dataY, canvas_widget_id, title="title", xtitle="X", ytitle="Y"):
        try:
            canvas_widget_id.layout().removeItem(canvas_widget_id.layout().itemAt(0))
        except:
            pass

        origin = (dataX[0], dataY[0])
        scale = (dataX[1] - dataX[0], dataY[1] - dataY[0])

        colormap = {"name": "temperature", "normalization": "linear",
                    "autoscale": True, "vmin": 0, "vmax": 0, "colors": 256}

        tmp = Plot2D()
        tmp.resetZoom()
        tmp.setXAxisAutoScale(True)
        tmp.setYAxisAutoScale(True)
        tmp.setGraphGrid(False)
        tmp.setKeepDataAspectRatio(True)
        tmp.yAxisInvertedAction.setVisible(False)
        tmp.setXAxisLogarithmic(False)
        tmp.setYAxisLogarithmic(False)
        tmp.getMaskAction().setVisible(False)
        tmp.getRoiAction().setVisible(False)
        tmp.getColormapAction().setVisible(True)
        tmp.setKeepDataAspectRatio(False)
        tmp.addImage(data2D.T,legend="1",scale=scale,origin=origin,colormap=colormap,replace=True)
        tmp.setActiveImage("1")
        tmp.setGraphXLabel(xtitle)
        tmp.setGraphYLabel(ytitle)
        tmp.setGraphTitle(title)

        canvas_widget_id.layout().addWidget(tmp)



if __name__ == "__main__":

    app = QApplication(sys.argv)
    w = OWConic()
    w.show()
    app.exec()
    w.saveSettings()

    # x = numpy.linspace(-10e-3, 10e-3, 101)
    # y = numpy.linspace(-100e-3, 100e-3, 1001)
    #
    # Z, X, Y =  diaboloid_exact_point_to_segment(p=29.3,q=19.53,theta=4.5e-3,x=x,y=y,)
    # Z0, X, Y = diaboloid_approximated_point_to_segment(p=29.3, q=19.53, theta=4.5e-3, x=x, y=y, detrend=1)
    #
    # from srxraylib.plot.gol import plot_image, plot
    # plot_image((Z0) * 1e-6, x * 1e-3, y * 1e-3, xtitle="X/mm", ytitle="Y/mm", title="Z (approximated)/um", aspect="auto")
    # plot_image((Z) * 1e-6, x * 1e-3, y * 1e-3, xtitle="X/mm", ytitle="Y/mm", title="Z (exact)/um", aspect="auto")
    # plot_image((Z-Z0) * 1e-6, x * 1e-3, y * 1e-3, xtitle="X/mm", ytitle="Y/mm", title="Z (exact)-Z(approximated)/um", aspect="auto")
    #
    # ZZ0 = Z0[:, y.size//2]
    # ZZ  = Z[:, y.size//2]
    # plot(x, ZZ0 - ZZ0.min(),
    #      x,  ZZ - ZZ.min(),
    #      xtitle="X", ytitle="Z", legend=["Z (approximated)","Z (exact)"])
    #
    # ZZ0 = Z0[x.size//2, :]
    # ZZ  = Z[x.size//2, :]
    # plot(y,  ZZ0 - ZZ0.min(),
    #      y,   ZZ - ZZ.min(),
    #      xtitle="Y", ytitle="Z", legend=["Z (approximated)","Z (exact)"])


#
# implementation of generic equations for quadrics
# seee chapter 6 in https://www.oecd-nea.org/upload/docs/application/pdf/2020-10/penelope-2018__a_code_system_for_monte_carlo_simulation_of_electron_and_photon_transport.pdf
#

import numpy


#
# some tools
#


def reduced_quadric(kind='plane'):
    # I1 x^2 + I2 y^2 + I3 z^2 + I4 z + I5 = 0  (Eq 6.23 in Penelope Manual)

    if kind == 'plane':  # (Ibid, table 6.1
        return [0,0,0,1,-1]
    elif kind == 'pair of parallel planes':
        return [0,0,1,0,-1]
    elif kind == 'sphere':
        return [1,1,1,0,-1]
    elif kind == 'cylinder':
        return [1,1,0,0,-1]
    elif kind == 'hyperbolic cylinder':
        return [1,-1,0,0,-1]
    elif kind == 'hyperbolic cylinder 2':
        return [-1,1,0,0,-1]
    elif kind == 'cone':
        return [1,1,-1,0,0]
    elif kind == 'one sheet hyperboloid':
        return [1,1,-1,0,-1]
    elif kind == 'two sheet hyperboloid':
        return [1,1,-1,0,1]
    elif kind == 'paraboloid':
        return [1,1,0,-1,0]
    elif kind == 'parabolic cylinder':
        return [1,0,0,-1,0]
    elif kind == 'parabolic cylinder 2':
        return [0,1,0,-1,0]
    elif kind == 'hyperbolic paraboloid':
        return [1,-1,0,-1,0]
    elif kind == 'hyperbolic paraboloid 2':
        return [-1,1,0,-1,0]
    else:
        raise Exception('Invalid reduced-quadric name.')


def scale_reduced_quadric(reduced_quadric, xscale=1.0, yscale=1.0, zscale=1.0, return_list=True):
    # see ibid, eq 6.24
    out = reduced_quadric.copy()
    out[0] /= xscale**2
    out[1] /= yscale**2
    out[2] /= zscale**2
    out[3] /= zscale
    out[4] = float(out[4])

    if return_list:
        return out
    else: # return 3 matrices (ibid, eq 6.26)
        A33 = numpy.array( [[out[0],0,0],
                            [0,out[1],0],
                            [0,0,out[2]]])
        A03 = numpy.array( [[0,0,out[3]]])
        A00 = out[4]
        return A33,A03,A00

def expand_reduced_quadric(reduced_quadric_list):
    # returns the 10 coefficients of the generic quartic from the 5 coeffs of the reduced quartic

    QXX = reduced_quadric_list[0]
    QYY = reduced_quadric_list[1]
    QZZ = reduced_quadric_list[2]
    QXY = 0.0
    QYZ = 0.0
    QXZ = 0.0
    QX = 0.0
    QY = 0.0
    QZ = reduced_quadric_list[3]
    Q0 = reduced_quadric_list[4]

    return [QXX,QYY,QZZ,QXY,QYZ,QXZ,QX,QY,QZ,Q0]



def quartic_coefficients_matrices_to_list(A2,A1,A0, fix_zeros=False):
    AXX = A2[1-1,1-1]
    AXY = A2[1-1,2-1] + A2[2-1,1-1]
    AXZ = A2[1-1,3-1] + A2[3-1,1-1]
    AYY = A2[2-1,2-1]
    AYZ = A2[2-1,3-1] + A2[3-1,2-1]
    AZZ = A2[3-1,3-1]
    AX = A1[1-1]
    AY = A1[2-1]
    AZ = A1[3-1]
    out_list = [AXX, AYY, AZZ, AXY, AYZ, AXZ, AX, AY, AZ, A0]

    if fix_zeros:
        for i in range(len(out_list)):
            if numpy.abs(out_list[i]) < 1e-15:
                out_list[i] = 0.0

    return out_list

def quadric_coefficients_list_to_matrices(quartic_coefficients_list):

    AXX, AYY, AZZ, AXY, AYZ, AXZ, AX, AY, AZ, A0 = quartic_coefficients_list

    B2 = numpy.zeros((3, 3))
    # ibid eq. 6.18
    B2[1 - 1, 1 - 1] = AXX
    B2[1 - 1, 2 - 1] = 0.5 * AXY
    B2[1 - 1, 3 - 1] = 0.5 * AXZ
    B2[2 - 1, 1 - 1] = B2[1 - 1, 2 - 1]
    B2[2 - 1, 2 - 1] = AYY
    B2[2 - 1, 3 - 1] = 0.5 * AYZ
    B2[3 - 1, 1 - 1] = B2[1 - 1, 3 - 1]
    B2[3 - 1, 2 - 1] = B2[2 - 1, 3 - 1]
    B2[3 - 1, 3 - 1] = AZZ

    B1 = numpy.zeros((3))
    B1[1 - 1] = AX
    B1[2 - 1] = AY
    B1[3 - 1] = AZ

    B0 = A0

    return B2,B1,B0


def euler_rotation_matrix(omega,theta,phi, shortcut=False, fix_zeros=True):
    STHETA = numpy.sin(theta)
    CTHETA = numpy.cos(theta)
    SPHI = numpy.sin(phi)
    CPHI = numpy.cos(phi)
    SOMEGA = numpy.sin(omega)
    COMEGA = numpy.cos(omega)

    # ibidm eq, 6.9
    R = numpy.zeros((3,3))

    if shortcut:
        R[1-1,1-1] = 1
        R[1-1,2-1] = 0
        R[1-1,3-1] = 0
        R[2-1,1-1] = 0
        R[2-1,2-1] = CTHETA
        R[2-1,3-1] = -STHETA
        R[3-1,1-1] = 0
        R[3-1,2-1] = STHETA
        R[3-1,3-1] = CTHETA
    else:
        R[1-1,1-1] = CPHI*CTHETA*COMEGA-SPHI*SOMEGA
        R[1-1,2-1] = -CPHI*CTHETA*SOMEGA-SPHI*COMEGA
        R[1-1,3-1] = CPHI*STHETA
        R[2-1,1-1] = SPHI*CTHETA*COMEGA+CPHI*SOMEGA
        R[2-1,2-1] = -SPHI*CTHETA*SOMEGA+CPHI*COMEGA
        R[2-1,3-1] = SPHI*STHETA
        R[3-1,1-1] = -STHETA*COMEGA
        R[3-1,2-1] = STHETA*SOMEGA
        R[3-1,3-1] = CTHETA

    if fix_zeros:
        for i in range(3):
            for j in range(3):
                if numpy.abs(R[i,j]) < 1e-15:
                    R[i,j] = 0.0

    return R


#
# main routine for transforming the 10 conic coefficients
#


# translated from penelope cortran code
def rotate_and_shift_quartic(quartic_coefficients_list,
                             omega=0.0, theta=0.0, phi=0.0,
                             D=[0,0,0]):
#
# initial quartic
#
    B2 = numpy.zeros((3,3))

    AXX,AYY,AZZ,AXY,AYZ,AXZ,AX,AY,AZ,A0 = quartic_coefficients_list

    # ibid eq. 6.18
    B2[1-1,1-1] = AXX
    B2[1-1,2-1] = 0.5 * AXY
    B2[1-1,3-1] = 0.5 * AXZ
    B2[2-1,1-1] = B2[1-1,2-1]
    B2[2-1,2-1] = AYY
    B2[2-1,3-1] = 0.5 * AYZ
    B2[3-1,1-1] = B2[1-1,3-1]
    B2[3-1,2-1] = B2[2-1,3-1]
    B2[3-1,3-1] = AZZ

    B1 = numpy.zeros((3))
    B1[1-1] = AX
    B1[2-1] = AY
    B1[3-1] = AZ

    B0 = A0

    # so far, input [or scaled] F(x,y,z) can be written in matrix form (ibid, eq. 6.25)
    # notation: X=(x,y,z) ; B1=(AX,AY,AZ)
    # F = X' . (B2 X) + B1 . X + B0

    D1 = numpy.array(D)

    #
    #  ****  Rotation matrix.
    #
    R = euler_rotation_matrix(omega, theta, phi)

    A1 = numpy.zeros(3)
    A2 = numpy.zeros((3,3))
    for I in range(3):
        A1[I] = 0.0
        for J in range(3):
            A1[I] = A1[I] + R[I,J] * B1[J]
            A2[I,J] = 0.0
            for M in range(3):
                for K in range(3):
                    A2[I,J] = A2[I,J] + R[I,K] * B2[K,M] * R[J,M]


    #
    #  ****  Shifted-rotated quadric.
    #

    # 3rd equation 6.29 and complete the second equation
    for I in range(3):
        A2D = 0.0
        for J in range(3):
            A2D = A2D + A2[I,J] * D1[J]

        B1[I] = A1[I] - 2.0 * A2D
        B0 = B0 + D1[I] * (A2D - A1[I])


    AXX = A2[1-1,1-1]
    AXY = A2[1-1,2-1] + A2[2-1,1-1]
    AXZ = A2[1-1,3-1] + A2[3-1,1-1]
    AYY = A2[2-1,2-1]
    AYZ = A2[2-1,3-1] + A2[3-1,2-1]
    AZZ = A2[3-1,3-1]
    AX = B1[1-1]
    AY = B1[2-1]
    AZ = B1[3-1]
    A0 = B0

    transformed_coefficients = [AXX,AYY,AZZ,AXY,AYZ,AXZ,AX,AY,AZ,A0]
    for i in range(len(transformed_coefficients)):
        if numpy.abs(transformed_coefficients[i]) < 1e-15:
            transformed_coefficients[i] = 0.0

    return transformed_coefficients


# full matrix numpy implementation
def rotate_and_shift_quartic_NEW(quartic_coefficients_list,
                             omega=0.0, theta=0.0, phi=0.0,
                             D=[0.0,0.0,0.0]):
    #
    # initial quartic in matrix format
    #

    # so far, input [or scaled] F(x,y,z) can be written in matrix form (ibid, eq. 6.25)
    # notation: X=(x,y,z) ; B1=(AX,AY,AZ)
    # F = X' . (B2 X) + B1 . X + B0
    B2, B1, B0 = quadric_coefficients_list_to_matrices(quartic_coefficients_list)

    #
    #  ****  Rotation matrix.
    #
    R = euler_rotation_matrix(omega, theta, phi)

    #
    #  ****  Shifted-rotated quadric.
    #

    D1 = numpy.array(D)

    # first equation 6.29
    A2 = numpy.dot(R, numpy.dot(B2,R.T))

    # 2nd equation 6.29
    A1_first_term = numpy.dot(R, B1)
    A1 = A1_first_term - 2 * numpy.dot(A2,D1)

    # 3rd equation 6.29
    A0 = B0 + numpy.dot(D1.T,
                    (numpy.dot(A2, D1) - numpy.dot(R, B1)))

    transformed_coefficients = quartic_coefficients_matrices_to_list(A2,A1,A0,fix_zeros=True)
    return transformed_coefficients

def rotate_and_shift_quartic_MATHEMATICA(quartic_coefficients_list,
                             omega=0.0, theta=0.0, phi=0.0,
                             D=[0.0,0.0,0.0]):

    from numpy import sin as Sin
    from numpy import cos as Cos

    cxx, cyy, czz, cxy, cyz, cxz, cx, cy, cz, c0 = quartic_coefficients_list

    Amat = numpy.zeros((3,3))



    Amat[0,0] = cxx
    Amat[0,1] = (cxy*Cos(theta) - cxz*Sin(theta))/2.
    Amat[0,2] = (cxz*Cos(theta) + cxy*Sin(theta))/2.
    Amat[1,0] = (cxy*Cos(theta) - cxz*Sin(theta))/2.
    Amat[1,1] = cyy*Cos(theta)**2 - cyz*Cos(theta)*Sin(theta) + czz*Sin(theta)**2
    Amat[1,2] = (cyz*Cos(2*theta) + (cyy - czz)*Sin(2*theta))/2.
    Amat[2,0] = (cxz*Cos(theta) + cxy*Sin(theta))/2.
    Amat[2,1] = (cyz*Cos(2*theta) + (cyy - czz)*Sin(2*theta))/2.
    Amat[2,2] = czz*Cos(theta)**2 + cyz*Cos(theta)*Sin(theta) + cyy*Sin(theta)**2


    tx = D[0]
    ty = D[1]
    tz = D[2]

    Avec = numpy.zeros(3)


    Avec[0] = cx - 2*cxx*tx - (cxy*ty + cxz*tz)*Cos(theta) + (cxz*ty - cxy*tz)*Sin(theta)
    Avec[1] = -((cyy + czz)*ty) + (cy - cxy*tx)*Cos(theta) - (cyy*ty - czz*ty + cyz*tz)*Cos(2*theta) + (-cz + cxz*tx)*Sin(theta) + (cyz*ty - cyy*tz + czz*tz)*Sin(2*theta)
    Avec[2] = -((cyz*ty + 2*czz*tz)*Cos(theta)**2) + Sin(theta)*(cy - cxy*tx + (cyz*ty - 2*cyy*tz)*Sin(theta)) + Cos(theta)*(cz - cxz*tx - 2*(cyy*ty - czz*ty + cyz*tz)*Sin(theta))

    A0 = c0 + \
        tx*(-cx + cxx*tx + tz*((cxz*Cos(theta))/2. + (cxy*Sin(theta))/2.) + ty*((cxy*Cos(theta))/2. - (cxz*Sin(theta))/2.)) + \
        tz*(-(cz*Cos(theta)) - cy*Sin(theta) + tx*((cxz*Cos(theta))/2. + (cxy*Sin(theta))/2.) + \
        tz*(Sin(theta)*((cyz*Cos(theta))/2. + cyy*Sin(theta)) + Cos(theta)*(czz*Cos(theta) + (cyz*Sin(theta))/2.)) + \
        ty*(Sin(theta)*(cyy*Cos(theta) - (cyz*Sin(theta))/2.) + Cos(theta)*((cyz*Cos(theta))/2. - czz*Sin(theta)))) + \
        ty*(-(cy*Cos(theta)) + cz*Sin(theta) + tx*((cxy*Cos(theta))/2. - (cxz*Sin(theta))/2.) + \
        tz*(Cos(theta)*((cyz*Cos(theta))/2. + cyy*Sin(theta)) - Sin(theta)*(czz*Cos(theta) + (cyz*Sin(theta))/2.)) + \
        ty*(Cos(theta)*(cyy*Cos(theta) - (cyz*Sin(theta))/2.) - Sin(theta)*((cyz*Cos(theta))/2. - czz*Sin(theta))))

    print(">>>", Amat)
    print(">>>", Avec)
    print(">>>", A0)
    Alist = quartic_coefficients_matrices_to_list(Amat,Avec,A0, fix_zeros=True)
    print(">>>", Alist)
    return Alist

# full implementation from Mathematica code
def rotate_and_shift_quartic_MATHEMATICAFULLEULER(quartic_coefficients_list,
                             omega=0.0, theta=0.0, phi=0.0,
                             D=[0.0,0.0,0.0]):

    from numpy import sin as Sin
    from numpy import cos as Cos

    cxx, cyy, czz, cxy, cyz, cxz, cx, cy, cz, c0 = quartic_coefficients_list

    Amat = numpy.zeros((3,3))

    # this is for theta-rotation around X
    # Amat[0,0] = cxx
    # Amat[0,1] = (cxy*Cos(theta) - cxz*Sin(theta))/2.
    # Amat[0,2] = (cxz*Cos(theta) + cxy*Sin(theta))/2.
    # Amat[1,0] = (cxy*Cos(theta) - cxz*Sin(theta))/2.
    # Amat[1,1] = cyy*Cos(theta)**2 - cyz*Cos(theta)*Sin(theta) + czz*Sin(theta)**2
    # Amat[1,2] = (cyz*Cos(2*theta) + (cyy - czz)*Sin(2*theta))/2.
    # Amat[2,0] = (cxz*Cos(theta) + cxy*Sin(theta))/2.
    # Amat[2,1] = (cyz*Cos(2*theta) + (cyy - czz)*Sin(2*theta))/2.
    # Amat[2,2] = czz*Cos(theta)**2 + cyz*Cos(theta)*Sin(theta) + cyy*Sin(theta)**2

    # this includes the three euler angles
    Amat[0,0] = ((Cos(omega)*Cos(phi)*Cos(theta) - Sin(omega)*Sin(phi))*
        (Cos(phi)*Cos(theta)*(2*cxx*Cos(omega) - cxy*Sin(omega)) - (cxy*Cos(omega) + 2*cxx*Sin(omega))*Sin(phi) + cxz*Cos(phi)*Sin(theta)) +
        (Cos(phi)*Cos(theta)*Sin(omega) + Cos(omega)*Sin(phi))*(Cos(phi)*Cos(theta)*(-(cxy*Cos(omega)) + 2*cyy*Sin(omega)) + (2*cyy*Cos(omega) + cxy*Sin(omega))*Sin(phi) -
        cyz*Cos(phi)*Sin(theta)) + Cos(phi)*Sin(theta)*(Cos(phi)*Cos(theta)*(cxz*Cos(omega) - cyz*Sin(omega)) - (cyz*Cos(omega) + cxz*Sin(omega))*Sin(phi) + 2*czz*Cos(phi)*Sin(theta))
        )/2.

    Amat[0,1] = (Cos(2*omega)*(4*cxy*Cos(2*phi)*Cos(theta) + (cxx - cyy)*(3 + Cos(2*theta))*Sin(2*phi)) +
        4*Cos(2*phi)*((cxx - cyy)*Cos(theta)*Sin(2*omega) + (cyz*Cos(omega) + cxz*Sin(omega))*Sin(theta)) +
        Sin(2*phi)*(-(cxy*(3 + Cos(2*theta))*Sin(2*omega)) - 2*(cxx + cyy - 2*czz)*Sin(theta)**2 + 2*(cxz*Cos(omega) - cyz*Sin(omega))*Sin(2*theta)))/8.


    Amat[0,2] = (Cos(phi)*Cos(2*theta)*(cxz*Cos(omega) - cyz*Sin(omega)) - Cos(theta)*(cyz*Cos(omega) + cxz*Sin(omega))*Sin(phi) +
        (cxy*Cos(2*omega) + (cxx - cyy)*Sin(2*omega))*Sin(phi)*Sin(theta) - (Cos(phi)*(cxx + cyy - 2*czz + (cxx - cyy)*Cos(2*omega) - cxy*Sin(2*omega))*Sin(2*theta))/2.)/2.


    Amat[1,0] = (Cos(2*omega)*(4*cxy*Cos(2*phi)*Cos(theta) + (cxx - cyy)*(3 + Cos(2*theta))*Sin(2*phi)) +
        4*Cos(2*phi)*((cxx - cyy)*Cos(theta)*Sin(2*omega) + (cyz*Cos(omega) + cxz*Sin(omega))*Sin(theta)) +
        Sin(2*phi)*(-(cxy*(3 + Cos(2*theta))*Sin(2*omega)) - 2*(cxx + cyy - 2*czz)*Sin(theta)**2 + 2*(cxz*Cos(omega) - cyz*Sin(omega))*Sin(2*theta)))/8.


    Amat[1,1] = ((Cos(phi)*Sin(omega) + Cos(omega)*Cos(theta)*Sin(phi))*(Cos(phi)*(cxy*Cos(omega) + 2*cxx*Sin(omega)) + Cos(theta)*(2*cxx*Cos(omega) - cxy*Sin(omega))*Sin(phi) +
        cxz*Sin(phi)*Sin(theta)) + (Cos(omega)*Cos(phi) - Cos(theta)*Sin(omega)*Sin(phi))*
        (cxy*Cos(phi)*Sin(omega) + Cos(omega)*(2*cyy*Cos(phi) + cxy*Cos(theta)*Sin(phi)) + Sin(phi)*(-2*cyy*Cos(theta)*Sin(omega) + cyz*Sin(theta))) +
        Sin(phi)*Sin(theta)*(cxz*Cos(phi)*Sin(omega) + Cos(omega)*(cyz*Cos(phi) + cxz*Cos(theta)*Sin(phi)) + Sin(phi)*(-(cyz*Cos(theta)*Sin(omega)) + 2*czz*Sin(theta))))/2.


    Amat[1,2] = ((Cos(phi)*Sin(omega) + Cos(omega)*Cos(theta)*Sin(phi))*(cxz*Cos(theta) + (-2*cxx*Cos(omega) + cxy*Sin(omega))*Sin(theta)) +
        (Cos(omega)*Cos(phi) - Cos(theta)*Sin(omega)*Sin(phi))*(cyz*Cos(theta) + (-(cxy*Cos(omega)) + 2*cyy*Sin(omega))*Sin(theta)) +
        Sin(phi)*Sin(theta)*(2*czz*Cos(theta) + (-(cxz*Cos(omega)) + cyz*Sin(omega))*Sin(theta)))/2.


    Amat[2,0] = (Cos(phi)*Cos(2*theta)*(cxz*Cos(omega) - cyz*Sin(omega)) - Cos(theta)*(cyz*Cos(omega) + cxz*Sin(omega))*Sin(phi) +
        (cxy*Cos(2*omega) + (cxx - cyy)*Sin(2*omega))*Sin(phi)*Sin(theta) - (Cos(phi)*(cxx + cyy - 2*czz + (cxx - cyy)*Cos(2*omega) - cxy*Sin(2*omega))*Sin(2*theta))/2.)/2.


    Amat[2,1] = (Cos(phi)*Cos(theta)*(cyz*Cos(omega) + cxz*Sin(omega)) + Cos(2*theta)*(cxz*Cos(omega) - cyz*Sin(omega))*Sin(phi) -
        Cos(phi)*(cxy*Cos(2*omega) + (cxx - cyy)*Sin(2*omega))*Sin(theta) - ((cxx + cyy - 2*czz + (cxx - cyy)*Cos(2*omega) - cxy*Sin(2*omega))*Sin(phi)*Sin(2*theta))/2.)/2.


    Amat[2,2] = czz*Cos(theta)**2 + Cos(theta)*(-(cxz*Cos(omega)) + cyz*Sin(omega))*Sin(theta) + (cxx*Cos(omega)**2 - cxy*Cos(omega)*Sin(omega) + cyy*Sin(omega)**2)*Sin(theta)**2


    tx = D[0]
    ty = D[1]
    tz = D[2]

    Avec = numpy.zeros(3)


    Avec[0] = cx - 2*cxx*tx - (cxy*ty + cxz*tz)*Cos(theta) + (cxz*ty - cxy*tz)*Sin(theta)
    Avec[1] = -((cyy + czz)*ty) + (cy - cxy*tx)*Cos(theta) - (cyy*ty - czz*ty + cyz*tz)*Cos(2*theta) + (-cz + cxz*tx)*Sin(theta) + (cyz*ty - cyy*tz + czz*tz)*Sin(2*theta)
    Avec[2] = -((cyz*ty + 2*czz*tz)*Cos(theta)**2) + Sin(theta)*(cy - cxy*tx + (cyz*ty - 2*cyy*tz)*Sin(theta)) + Cos(theta)*(cz - cxz*tx - 2*(cyy*ty - czz*ty + cyz*tz)*Sin(theta))

    A0 = c0 + \
        tx*(-cx + cxx*tx + tz*((cxz*Cos(theta))/2. + (cxy*Sin(theta))/2.) + ty*((cxy*Cos(theta))/2. - (cxz*Sin(theta))/2.)) + \
        tz*(-(cz*Cos(theta)) - cy*Sin(theta) + tx*((cxz*Cos(theta))/2. + (cxy*Sin(theta))/2.) + \
        tz*(Sin(theta)*((cyz*Cos(theta))/2. + cyy*Sin(theta)) + Cos(theta)*(czz*Cos(theta) + (cyz*Sin(theta))/2.)) + \
        ty*(Sin(theta)*(cyy*Cos(theta) - (cyz*Sin(theta))/2.) + Cos(theta)*((cyz*Cos(theta))/2. - czz*Sin(theta)))) + \
        ty*(-(cy*Cos(theta)) + cz*Sin(theta) + tx*((cxy*Cos(theta))/2. - (cxz*Sin(theta))/2.) + \
        tz*(Cos(theta)*((cyz*Cos(theta))/2. + cyy*Sin(theta)) - Sin(theta)*(czz*Cos(theta) + (cyz*Sin(theta))/2.)) + \
        ty*(Cos(theta)*(cyy*Cos(theta) - (cyz*Sin(theta))/2.) - Sin(theta)*((cyz*Cos(theta))/2. - czz*Sin(theta))))

    Alist = quartic_coefficients_matrices_to_list(Amat,Avec,A0, fix_zeros=True)

    # print(">>>", Amat)
    # print(">>>", Avec)
    # print(">>>", A0)
    # print(">>>", Alist)
    return Alist

#
# specific conics
#

def sphere(ssour=10,simag=3,theta_grazing=3e-3):
    theta = (numpy.pi / 2) - theta_grazing
    rmirr = ssour * simag * 2 / numpy.cos(theta) / (ssour + simag)

    s1 = reduced_quadric('sphere')
    s2 = scale_reduced_quadric(s1, xscale=rmirr, yscale=rmirr, zscale=rmirr, return_list=True)
    s3 = expand_reduced_quadric(s2)
    s4 = rotate_and_shift_quartic(s3,
                             omega=0.0, theta=0.0, phi=0.0,
                             D=[0.0,0.0,rmirr])
    s5 = s4.copy()
    for i in range(10):
        s5[i] /= s4[0]

    print("Sphere: ")
    print("   R: ", rmirr)
    print("   reduced: ", s1)
    print("   scaled: ", s2)
    print("   expanded: ", s3)
    print("   rotated and shifted: ", s4)
    print("   normalized: ", s5)
    return {'p':ssour, 'q':simag, 'theta_grazing':theta_grazing, 'radius':rmirr, 'ccc':s5}

def paraboloid(ssour=10,simag=3,theta_grazing=3e-3, verbose=True):

    if ssour >= simag: # focusing
        a = simag * numpy.sin(theta_grazing)**2
        YCEN = - simag * numpy.sin(2 * theta_grazing)
        ZCEN = simag * numpy.cos(theta_grazing)**2

        CENTER = numpy.array([0, YCEN, ZCEN])

        NORMAL = numpy.array((0, -2 * YCEN, 4 * a))
        NORMAL_MOD = numpy.sqrt(NORMAL[0] ** 2 + NORMAL[1] ** 2 + NORMAL[2] ** 2)
        NORMAL /= NORMAL_MOD
        NORMAL_NEW = numpy.array((0, numpy.cos(theta_grazing), numpy.sin(theta_grazing)))
        # Euler angles
        omega = 1 / 2 * numpy.pi
        theta = numpy.pi / 2 - theta_grazing
        phi = 3 / 2 * numpy.pi
    else: # collimating
        a = ssour * numpy.sin(theta_grazing)**2
        YCEN = + ssour * numpy.sin(2 * theta_grazing)
        ZCEN = ssour * numpy.cos(theta_grazing)**2

        CENTER = numpy.array([0, YCEN, ZCEN])

        NORMAL = numpy.array((0, -2 * YCEN, 4 * a))
        NORMAL_MOD = numpy.sqrt(NORMAL[0] ** 2 + NORMAL[1] ** 2 + NORMAL[2] ** 2)
        NORMAL /= NORMAL_MOD
        NORMAL_NEW = numpy.array((0, -numpy.cos(theta_grazing), numpy.sin(theta_grazing)))
        # Euler angles
        omega = 1 / 2 * numpy.pi
        theta = -(numpy.pi / 2 - theta_grazing) #+ numpy.pi
        phi = 3 / 2 * numpy.pi


    PARAM = 2 * a
    if verbose:
        txt = ""
        if ssour >= simag:
            txt += "** Source is at infinity\n"
            txt += "** q=%f, theta_grazing=%f rad, theta_normal=%f rad\n" % (simag, theta_grazing, theta)
        else:
            txt += "** Image is at infinity\n"
            txt += "** p=%f, theta_grazing=%f rad, theta_normal=%f rad\n" % (ssour, theta_grazing, theta)
        txt += '** Parabloid a=%f \n' % a
        txt += '** Parabloid of revolution PARAM=%f \n' % PARAM
        txt += '** 4 * a=%f \n' % (4 * a)
        txt += '** Optical element center at: (%f,%f,%f)\n' % (CENTER[0],CENTER[1],CENTER[2])
        txt += '** Normal: (%f,%f,%f) \n' % (NORMAL[0],NORMAL[1],NORMAL[2])
        txt += '** Normal NEW!!: (%f,%f,%f) \n' % (NORMAL_NEW[0],NORMAL_NEW[1],NORMAL_NEW[2])
        print(txt)



    ROTATED_CENTER = numpy.dot(euler_rotation_matrix(omega, theta, phi), CENTER)

    s1 = reduced_quadric('paraboloid')
    s2 = scale_reduced_quadric(s1, xscale=1.0, yscale=1.0, zscale=(1/4/a), return_list=True)
    s3 = expand_reduced_quadric(s2)

    s4 = rotate_and_shift_quartic_NEW(s3,
                                      omega=omega, theta=theta, phi=phi,
                                      D=-ROTATED_CENTER)

    s5 = s4.copy()
    # for i in range(10):
    #     s5[i] /= s4[0]

    print("**Paraboloid: ")
    print("**   a, theta[deg]: ", PARAM, theta*180/numpy.pi)
    print("**   reduced: ", s1)
    print("**   scaled: ", s2)
    print("**   expanded: ", s3)
    print("**   rotated and shifted: ", s4)
    print("**   normalized: ", s5)
    return {'p':ssour, 'q':simag, 'theta_grazing':theta_grazing, 'a':a, 'center':CENTER, 'normal':NORMAL,  'ccc':s5}


def ellipsoid(ssour=10, simag=3, theta_grazing=3e-3, verbose=True):

    a = 0.5 * (ssour + simag)
    b = numpy.sqrt(ssour * simag) * numpy.sin(theta_grazing)
    c = numpy.sqrt(a**2 - b**2)

    YCEN = (ssour**2 - simag**2) / 4 / c
    ZCEN = -b * numpy.sqrt(1 - YCEN**2 / a**2)
    CENTER = numpy.array([0,YCEN, ZCEN])

    NORMAL = numpy.array((0, -2 * YCEN / a**2, -2 * ZCEN / b**2))
    NORMAL_MOD = numpy.sqrt(NORMAL[0]**2 + NORMAL[1]**2 + NORMAL[2]**2)
    NORMAL /= NORMAL_MOD

    AXMAJ = a
    AXMIN = b
    AFOCI = c
    ECCENT = c / a

    #
    # Euler angles
    #
    omega = 1/2 * numpy.pi
    theta = numpy.arcsin(NORMAL[1]) # -theta_grazing # numpy.arccos(RNCEN[3-1])
    phi = 3/2 * numpy.pi

    if verbose:
        txt = ""
        txt += "** p=%f, q=%f, theta_grazing=%f rad, theta_normal=%f rad\n" % (ssour, simag, theta_grazing, (numpy.pi / 2) - theta_grazing)
        txt += '** Ellipsoid of revolution a=%f \n' % AXMAJ
        txt += '** Ellipsoid of revolution b=%f \n' % AXMIN
        txt += '** Ellipsoid of revolution 1/a**2=%f \n' % (1/AXMAJ**2)
        txt += '** Ellipsoid of revolution 1/b**2=%f \n' % (1/AXMIN**2)
        txt += '** Ellipsoid of revolution c=sqrt(a^2-b^2)=%f \n' % AFOCI
        txt += '** Ellipsoid of revolution focal distance c^2=%f \n' % (AFOCI ** 2)
        txt += '** Ellipsoid of revolution excentricity: %f \n' % ECCENT
        txt += '** Optical element center at: [%f,%f,%f]\n' % (CENTER[0], CENTER[1], CENTER[2])
        txt += '** Optical element normal: [%f,%f,%f]\n' % (NORMAL[0], NORMAL[1], NORMAL[2])
        txt += '** Optical element tangent: [%f,%f,%f]\n' % (NORMAL[0], NORMAL[2], -NORMAL[1])
        txt += '** THETA from NORMAL %f deg\n' % (numpy.arcsin(NORMAL[1]) * 180 / numpy.pi)
        txt += '** THETA from NORMAL %f deg\n' % (numpy.arccos(NORMAL[2]) * 180 / numpy.pi)
        txt += '** THETA from EULER %f deg\n' % (theta * 180 / numpy.pi)
        txt += '** B nz ycen - A ny zcen: %f\n' % ((1/a**2) *NORMAL[2] * YCEN - (1/b**2) * NORMAL[1] * ZCEN)
        print(txt)

        print(txt)



    s1 = reduced_quadric('sphere')
    s2 = scale_reduced_quadric(s1, xscale=AXMIN, yscale=AXMAJ, zscale=AXMIN, return_list=True)
    s3 = expand_reduced_quadric(s2)
    D=-numpy.dot(euler_rotation_matrix(omega, theta, phi), CENTER)
    s4 = rotate_and_shift_quartic_NEW(s3,
                             omega=omega, theta=theta, phi=phi,
                             D=-numpy.dot(euler_rotation_matrix(omega, theta, phi), CENTER))
    s5 = s4.copy()
    # for i in range(10):
    #     s5[i] /= s4[0]

    print("**Ellipsoid: ")
    print("**   a,b, theta_grazing[rad]: ", AXMAJ, AXMIN, theta_grazing)
    print("**   euler [deg]: ", omega * 180 / numpy.pi, theta * 180 / numpy.pi, phi * 180 / numpy.pi)
    print("**   D: ", D[0],D[1],D[2])
    print("**   reduced: ", s1)
    print("**   scaled: ", s2)
    print("**   expanded: ", s3)
    print("**   rotated and shifted: ", s4)
    print("**   normalized: ", s5)

    return {'p':ssour, 'q':simag, 'theta_grazing':theta_grazing,
                   'a':a, 'b':b, 'c':c,
                   'center':CENTER, 'normal':NORMAL,  'ccc':s5}


def hyperboloid(ssour=10, simag=3, theta_grazing=3e-3, verbose=True):

    a = 0.5 * numpy.abs(ssour - simag)
    c = 0.5 * numpy.sqrt(ssour**2 + simag**2 - 2 * ssour * simag * numpy.cos(2 * theta_grazing))
    b = numpy.sqrt(c**2 - a**2)

    # Large p (p>q)
    if ssour > simag: # select center in first quadrant
        YCEN = (ssour**2 - simag**2) / 4 / c
        ZCEN = b * numpy.sqrt(YCEN**2 / a**2 - 1)
        NORMAL = numpy.array((0, -2 * YCEN / a ** 2, 2 * ZCEN / b ** 2)) #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        NORMAL_MOD = numpy.sqrt(NORMAL[0] ** 2 + NORMAL[1] ** 2 + NORMAL[2] ** 2)
        NORMAL /= NORMAL_MOD
        #
        # Euler angles
        #
        omega = 1 / 2 * numpy.pi
        theta = numpy.arcsin(NORMAL[1]) #<<<<<<<<<<<<<<<<<<<<<
        phi = 3 / 2 * numpy.pi
    # Large q
    else: # center in 2nd quadrant
        YCEN = (ssour**2 - simag**2) / 4 / c
        ZCEN = b * numpy.sqrt(YCEN**2 / a**2 - 1)
        NORMAL = -numpy.array((0, -2 * YCEN / a ** 2, 2 * ZCEN / b ** 2)) #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        NORMAL_MOD = numpy.sqrt(NORMAL[0] ** 2 + NORMAL[1] ** 2 + NORMAL[2] ** 2)
        NORMAL /= NORMAL_MOD
        #
        # Euler angles
        #
        omega = 1 / 2 * numpy.pi
        theta = -numpy.arccos(NORMAL[2]) #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        phi = 3 / 2 * numpy.pi

    CENTER = numpy.array([0,YCEN, ZCEN])

    ecc = c / a


    # #
    # # Euler angles
    # #
    # omega = 1/2 * numpy.pi
    # theta = numpy.arcsin(NORMAL[1])# numpy.pi - numpy.abs(numpy.arcsin(NORMAL[1])) # numpy.arcsin(NORMAL[1]) # -theta_grazing # numpy.arccos(RNCEN[3-1])
    # phi = 3/2 * numpy.pi

    if verbose:
        txt = ""
        txt += "** p=%f, q=%f, theta_grazing=%f rad, theta_normal=%f rad\n" % (ssour, simag, theta_grazing, (numpy.pi / 2) - theta_grazing)
        txt += '** Hyperboloid of revolution a=%f \n' % a
        txt += '** Hyperboloid of revolution b=%f \n' % b
        txt += '** Hyperboloid of revolution 1/a**2=%f \n' % (1/a**2)
        txt += '** Hyperboloid of revolution 1/b**2=%f \n' % (1/b**2)
        txt += '** Hyperboloid of revolution c=sqrt(a^2+b^2)=%f \n' % c
        txt += '** Hyperboloid of revolution focal distance c^2=%f \n' % (c ** 2)
        txt += '** Hyperboloid of revolution excentricity: %f \n' % ecc
        txt += '** Optical element center at: [%f,%f,%f]\n' % (CENTER[0], CENTER[1], CENTER[2])
        txt += '** Optical element normal: [%f,%f,%f]\n' % (NORMAL[0], NORMAL[1], NORMAL[2])
        txt += '** Optical element tangent: [%f,%f,%f]\n' % (NORMAL[0], NORMAL[2], -NORMAL[1])
        txt += '** THETA from NORMAL %f rad\n' % (numpy.arcsin(NORMAL[1]) )
        txt += '** THETA from NORMAL %f rad\n' % (numpy.arccos(NORMAL[2]) )
        txt += '** THETA from euler %f rad\n' % (theta )
        # txt += '** B nz ycen - A ny zcen: %f\n' % ((1/a**2) *NORMAL[2] * YCEN - (1/b**2) * NORMAL[1] * ZCEN)
        print(txt)


    s1 = [-1,1,-1,0,-1] # reduced_quadric('one sheet hyperboloid')
    s2 = scale_reduced_quadric(s1, xscale=b, yscale=a, zscale=b, return_list=True)
    s3 = expand_reduced_quadric(s2)

    D = -numpy.dot(euler_rotation_matrix(omega, theta, phi), CENTER)
    # D = numpy.dot(euler_rotation_matrix(0, numpy.pi, 0.0), D) # make the incident beam in the negative part

    s4 = rotate_and_shift_quartic_NEW(s3,
                             omega=omega, theta=theta, phi=phi,
                             D=D)
    # if ssour < simag:
    #     # make the incident beam in the negative part
    #     s4 = rotate_and_shift_quartic_NEW(s4,
    #                              omega=numpy.pi/2, theta=numpy.pi, phi=numpy.pi*3/2,
    #                              D=[0,0,0])
    s5 = s4.copy()
    # for i in range(10):
    #     s5[i] /= s4[0]

    if verbose:
        print("**Hyperboloid: ")
        print("**   a,b, theta_grazing[rad]: ", a, b, theta_grazing)
        print("**   euler [deg]: ", omega * 180 / numpy.pi, theta * 180 / numpy.pi, phi * 180 / numpy.pi)
        print("**   D: ", D[0],D[1],D[2])
        print("**   rotated N: ", numpy.dot(euler_rotation_matrix(omega, theta, phi), NORMAL))
        print("**   reduced: ", s1)
        print("**   scaled: ", s2)
        print("**   expanded: ", s3)
        print("**   rotated and shifted: ", s4)
        print("**   normalized: ", s5)
    A = -1/b**2
    B = 1/a**2
    ny = NORMAL[1]
    nz = NORMAL[2]
    ccc = [A, A*ny**2+B*nz**2,A*nz**2+B*ny**2,0,2*(B-A)*ny*nz,0.,0.,0.,2*(B*ny*YCEN+A*nz*ZCEN),0.]
    if verbose: print("**   using SHADOW way: ", ccc)

    return {'p':ssour, 'q':simag, 'theta_grazing':theta_grazing,
                   'a':a, 'b':b, 'c':c,
                   'center':CENTER, 'normal':NORMAL,  'ccc':s5}
#
# TESTING ROUTINES
#
def sphere_check():
    s5 = sphere(ssour=10,simag=3,theta_grazing=3e-3)
    c=numpy.zeros(10)
    c[1-1]= 1
    c[2-1]= 1
    c[3-1]= 1
    c[4-1]= 0
    c[5-1]= 0
    c[6-1]= 0
    c[7-1]= 0
    c[8-1]= 0
    c[9-1]= -3076.93
    c[10-1]= 0

    for i in range(10):
        print(s5['ccc'][i] , c[i])
        assert(numpy.abs(s5['ccc'][i] - c[i]) < 1e-2)

def ellipsoid_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=False):


    ccc = S4Conic.initialize_as_ellipsoid_from_focal_distances(ssour, simag, theta_grazing,
                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
    print("ccc: ", ccc.get_coefficients())

    s5 = ellipsoid(ssour=ssour,simag=simag,theta_grazing=theta_grazing)

    c = ccc.get_coefficients()
    print("ccc: ", c)
    print("s5: ", s5['ccc'])


    for i in range(10):
        print(i, c[i], s5['ccc'][i])
        assert(numpy.abs(s5['ccc'][i] - c[i]) < 1e-2)

    # view_conic(s5, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1)
    if do_plot:
        compare_conics(s5['ccc'], ccc.get_coefficients(), x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1,
                       titles=['s5','ccc'])

def hyperboloid_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=False):


    # ccc = S4Conic.initialize_as_hyperboloid_from_focal_distances(ssour, simag, theta_grazing,
    #                                     cylindrical=0, cylangle=0.0, switch_convexity=0)
    # print("ccc: ", ccc.get_coefficients())

    if ssour < simag:
        c = [-3703.714814855263, -0.033333333332666124, -3703.5998488688692, 0.0, 41.269717460237345, 0.0, 0.0, 3.0795921368564905e-12, 190.47647619130197, 0.0]
        branch = 0
    else:
        c = [-3703.714814855263, -0.03333333333333342, -3703.5998488688683, 0.0, -41.269717460357114, 0.0, 0.0, 0.0, -190.47647619130197, 0.0]
        branch = 0

    s5 = hyperboloid(ssour=ssour,simag=simag,theta_grazing=theta_grazing)

    print("ccc: ", c)
    print("s5: ", s5['ccc'])


    for i in range(10):
        print(i, c[i], s5['ccc'][i])
        assert(numpy.abs(s5['ccc'][i] - c[i]) < 1e-2)

    # view_conic(s5, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1)
    if do_plot:
        compare_conics(s5['ccc'], c, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1,
                       titles=['s5','ccc'], branch=branch)

    return s5

def parabola_check(ssour=10,simag=10,theta_grazing=3e-3, do_plot=False):


    ccc = S4Conic.initialize_as_paraboloid_from_focal_distances(ssour, simag, theta_grazing,
                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
    print("ccc: ", ccc.get_coefficients())

    s5 = paraboloid(ssour=ssour,simag=simag,theta_grazing=theta_grazing)

    print("ccc: ", ccc.get_coefficients())
    print("s5: ", s5['ccc'])

    c = ccc.get_coefficients()
    for i in range(10):
        print(i, c[i], s5['ccc'][i])

    for i in range(10):
        print(s5['ccc'][i] , c[i])
        assert(numpy.abs(s5['ccc'][i] - c[i]) < 1e-2)

    # view_conic(s5, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1)
    if do_plot:
        compare_conics(s5, ccc.get_coefficients(), x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1,
                       titles=['s5','ccc'])


def height(ccc,y=0,x=0,return_solution=0):
    """

    :param y: a scalar, vector or mesh
    :param x: a scalar, vector or mesh
        y and x must be homogeneous, otherwise an error will occur:
         both scalars
         both mesh
         one scalar and another vector
    :param return_solution: 0 = guess the solution with zero at pole,
                            1 = get first solution
                            2 = get second solution
    :return: the height scalar/vector/mesh depending on inputs
    """
    aa = ccc[2]
    bb = ccc[4] * y + ccc[5] * x + ccc[8]
    cc = ccc[0] * x**2 + ccc[1] * y**2 + ccc[3] * x * y + \
        ccc[6] * x + ccc[7] * y + ccc[9]

    if aa != 0:
        discr = bb**2 - 4 * aa * cc + 0j
        s1 = (-bb + numpy.sqrt(discr)) / 2 / aa
        s2 = (-bb - numpy.sqrt(discr)) / 2 / aa

        if return_solution == 0: # select the solution close to zero at pole
            if numpy.abs(s1).min() < numpy.abs(s2).min():
                ss = s1
            else:
                ss = s2
        elif return_solution == 1:
            ss = s1
        else:
            ss = s2
    else:
        ss = -cc / bb

    return numpy.real(ss)

if __name__ == "__main__":

    try:
        from shadow4.optical_surfaces.s4_conic import S4Conic  # for comparison
        from shadow4tests.devel.wolter.conic_viewer import view_conic, compare_conics  # for plot
    except:
        pass

    sphere_check()

    parabola_check(ssour=1e8,simag=10,theta_grazing=3e-3, do_plot=0)
    parabola_check(ssour=10,simag=1e9,theta_grazing=3e-3, do_plot=0)
    #
    ellipsoid_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=0)
    #
    #
    hyperboloid_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=0)
    hyperboloid_check(ssour=3, simag=10, theta_grazing=3e-3, do_plot=0)





from orangewidget.settings import Setting
from orangewidget import gui

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import write_surface_file

from orangecontrib.esrf.syned.util.lens import Lens # TODO: from syned.beamline.optical_elements....

from orangecontrib.esrf.wofry.util.lens import WOLens # from wofryimpl...
from orangecontrib.wofry.widgets.gui.ow_optical_element import OWWOOpticalElement


class OWWORealLens2D(OWWOOpticalElement):

    name = "Real Lens 2D"
    id = "WofryLens2D"
    description = "Wofry: Real Lens 2D"
    icon = "icons/lens2d.png"
    priority = 204

    number_of_curved_surfaces = Setting(2)
    lens_radius = Setting(200e-6)

    surface_shape = Setting(0)
    two_d_lens = Setting(0)
    wall_thickness = Setting(10e-6)

    n_lenses = Setting(1)
    material = Setting(1)
    refraction_index_delta = Setting(5.3e-7)
    att_coefficient = Setting(0.00357382)

    aperture_shape = Setting(0)
    aperture_dimension_v = Setting(100e-6)
    aperture_dimension_h = Setting(200e-6)

    shape = Setting(1)
    radius = Setting(100e-6)

    write_profile_flag = Setting(0)
    write_profile = Setting("lens_profile_2D.h5")

    def __init__(self):
        super().__init__(is_automatic=True, show_view_options=True, show_script_tab=True)

    def set_visible(self):
        self.lens_radius_box_id.setVisible(self.number_of_curved_surfaces != 0)
        self.lens_width_id.setVisible(self.aperture_shape == 1)
        self.box_file_out.setVisible(self.write_profile_flag == 1)
        self.box_refraction_index_id.setVisible(self.material in [0])
        self.box_att_coefficient_id.setVisible(self.material in [0])

    def get_material_name(self, index=None):
        materials_list = ["External", "Be", "Al", "Diamond"]
        if index is None:
            return materials_list
        else:
            return materials_list[index]

    def draw_specific_box(self):

        self.lens_box = oasysgui.widgetBox(self.tab_bas, "Real Lens Setting", addSpace=False, orientation="vertical",
                                           height=350)

        oasysgui.lineEdit(self.lens_box, self, "wall_thickness", "(t_wall) Wall thickness [m]", labelWidth=260, valueType=float, orientation="horizontal")


        oasysgui.lineEdit(self.lens_box, self, "n_lenses", "Number of lenses",
                          labelWidth=300, valueType=int, orientation="horizontal")

        gui.comboBox(self.lens_box, self, "material", label="Lens material",
                     items=self.get_material_name(),callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")

        self.box_refraction_index_id = oasysgui.widgetBox(self.lens_box, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_refraction_index_id, self, "refraction_index_delta", "Refraction index delta",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("refraction_index_delta")

        self.box_att_coefficient_id = oasysgui.widgetBox(self.lens_box, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_att_coefficient_id, self, "att_coefficient", "Attenuation coefficient [m-1]",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("att_coefficient")



        gui.comboBox(self.lens_box, self, "number_of_curved_surfaces", label="Number of curved surfaces", labelWidth=350,
                     items=["0 (parallel plate)", "1 (plano-concave)", "2 (bi-concave)"],
                     sendSelectedValue=False, orientation="horizontal", callback=self.set_visible)


        self.lens_radius_box_id = oasysgui.widgetBox(self.lens_box, orientation="vertical", height=None)
        oasysgui.lineEdit(self.lens_radius_box_id, self, "lens_radius", "(R) radius of curvature [m]", labelWidth=260,
                          valueType=float, orientation="horizontal",)

        gui.comboBox(self.lens_radius_box_id, self, "surface_shape", label="Lens shape", labelWidth=350,
                     items=["Parabolic", "Circular"],
                     sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(self.lens_radius_box_id, self, "two_d_lens", label="Focusing in", labelWidth=350,
                     items=["2D", "1D (tangential)", "1D (sagittal)"], sendSelectedValue=False, orientation="horizontal")


        # super().draw_specific_box() # adds boundary box
        gui.comboBox(self.lens_box, self, "aperture_shape", label="Aperture shape", labelWidth=350,
                     items=["Circular", "Rectangular"],
                     sendSelectedValue=False, orientation="horizontal", callback=self.set_visible)
        oasysgui.lineEdit(self.lens_box, self, "aperture_dimension_v", "Aperture V (height/diameter) [m]", labelWidth=260,
                          valueType=float, orientation="horizontal",)

        self.lens_width_id = oasysgui.widgetBox(self.lens_box, orientation="vertical", height=None)
        oasysgui.lineEdit(self.lens_width_id, self, "aperture_dimension_h", "Aperture H (width) [m]", labelWidth=260,
                          valueType=float, orientation="horizontal",)


        # files i/o tab
        self.tab_files = oasysgui.createTabPage(self.tabs_setting, "File I/O")
        files_box = oasysgui.widgetBox(self.tab_files, "Files", addSpace=True, orientation="vertical")


        gui.comboBox(files_box, self, "write_profile_flag", label="Dump profile to file",
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal",
                     callback=self.set_visible)

        self.box_file_out = gui.widgetBox(files_box, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.box_file_out, self, "write_profile", "File name",
                            labelWidth=200, valueType=str, orientation="horizontal")


        self.set_visible()

    def get_optical_element(self):

        return WOLens.create_from_keywords(
            name=self.name,
            number_of_curved_surfaces=self.number_of_curved_surfaces,
            two_d_lens=self.two_d_lens,
            surface_shape=self.surface_shape,
            wall_thickness=self.wall_thickness,
            lens_radius=self.lens_radius,
            material=self.get_material_name(index=self.material),
            refraction_index_delta=self.refraction_index_delta,
            att_coefficient=self.att_coefficient,
            n_lenses=self.n_lenses,
            aperture_shape=self.aperture_shape,
            aperture_dimension_h=self.aperture_dimension_h,
            aperture_dimension_v=self.aperture_dimension_v,
        )

    def check_data(self):
        super().check_data()
        # congruence.checkStrictlyPositiveNumber(numpy.abs(self.focal_x), "Horizontal Focal Length")
        # congruence.checkStrictlyPositiveNumber(numpy.abs(self.focal_y), "Vertical Focal Length")

    def receive_specific_syned_data(self, optical_element):
        if not optical_element is None:
            if isinstance(optical_element, Lens):
                self.lens_radius = optical_element._radius
                self.wall_thickness = optical_element._thickness
                self.material = optical_element._material
            else:
                raise Exception("Syned Data not correct: Optical Element is not a Lens")
        else:
            raise Exception("Syned Data not correct: Empty Optical Element")

    # overwrite this method to add tab with thickness profile

    def initializeTabs(self): #TODO Customize parent and remove (like in 1D)
        size = len(self.tab)
        indexes = range(0, size)

        for index in indexes:
            self.tabs.removeTab(size-1-index)

        titles = ["Intensity","Phase","Thickness profile"]
        self.tab = []
        self.plot_canvas = []

        for index in range(0, len(titles)):
            self.tab.append(gui.createTabPage(self.tabs, titles[index]))
            self.plot_canvas.append(None)

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

    def propagate_wavefront(self):
        super().propagate_wavefront()

        if self.write_profile_flag == 1:
            xx, yy, s = self.get_optical_element().get_surface_thickness_mesh(self.input_data.get_wavefront())
            write_surface_file(s.T, xx, yy, self.write_profile, overwrite=True)
            print("\nFile for OASYS " + self.write_profile + " written to disk.")

    def do_plot_results(self, progressBarValue=80): # OVERWRITTEN

        super().do_plot_results(progressBarValue)
        if not self.view_type == 0:
            if not self.wavefront_to_plot is None:

                self.progressBarSet(progressBarValue)

                x, y, lens_thickness = self.get_optical_element().get_surface_thickness_mesh(self.input_data.get_wavefront())

                self.plot_data2D(data2D=1e6*lens_thickness,
                             dataX=1e6*x,
                             dataY=1e6*y,
                             progressBarValue=progressBarValue,
                             tabs_canvas_index=2,
                             plot_canvas_index=2,
                             title="O.E. Thickness profile in $\mu$m",
                             xtitle="Horizontal [$\mu$m] ( %d pixels)"%(self.wavefront_to_plot.get_coordinate_x().size),
                             ytitle="Vertical [$\mu$m] ( %d pixels)"%(self.wavefront_to_plot.get_coordinate_y().size))

                self.progressBarFinished()

if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    def get_example_wofry_data():
        from wofryimpl.propagator.light_source import WOLightSource
        from wofryimpl.beamline.beamline import WOBeamline
        from orangecontrib.wofry.util.wofry_objects import WofryData

        light_source = WOLightSource(dimension=2,
                                     initialize_from=0,
                                     range_from_h=-0.002,
                                     range_to_h=0.002,
                                     range_from_v=-0.001,
                                     range_to_v=0.001,
                                     number_of_points_h=400,
                                     number_of_points_v=200,
                                     energy=10000.0,
                                     )

        return WofryData(wavefront=light_source.get_wavefront(),
                           beamline=WOBeamline(light_source=light_source))


    a = QApplication(sys.argv)
    ow = OWWORealLens2D()
    ow.set_input(get_example_wofry_data())
    ow.show()
    a.exec_()
    ow.saveSettings()



import sys, os, platform
from PyQt5.QtWidgets import QApplication

from orangewidget import gui
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui, congruence

from orangecontrib.xoppy.widgets.gui.ow_xoppy_widget import XoppyWidget

from oasys.widgets.exchange import DataExchangeObject

import numpy
import scipy.constants as codata

from syned.widget.widget_decorator import WidgetDecorator
import syned.beamline.beamline as synedb
import syned.storage_ring.magnetic_structures.insertion_device as synedid

from xoppylib.xoppy_run_binaries import xoppy_calc_ws

class OWID19ws(XoppyWidget,WidgetDecorator):
    name = "ID19WS"
    id = "orange.widgets.dataws"
    description = "Wiggler Spectrum on a Screen"
    icon = "icons/ws_id19_hpw150.png"
    priority = 1
    category = ""
    keywords = ["xoppy", "ws"]

    ENERGY = Setting(6.0)
    CUR = Setting(200.0)
    PERIOD = Setting(15.0)
    N = Setting(28)
    KX = Setting(0.0)
    KY = Setting(19)
    GAP = Setting(30)
    EMIN = Setting(1000.0)
    EMAX = Setting(200000.0)
    NEE = Setting(500)
    D = Setting(30.0)
    XPC = Setting(0.0)
    YPC = Setting(0.0)
    XPS = Setting(2.0)
    YPS = Setting(2.0)
    NXP = Setting(10)
    NYP = Setting(10)

    inputs = WidgetDecorator.syned_input_data()

    def __init__(self):
        super().__init__(show_script_tab=True)
    
    def build_gui(self):

        box = oasysgui.widgetBox(self.controlArea, self.name + " Input Parameters", orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)

        idx = -1 


        #widget index 1
        idx += 1 
        box1 = gui.widgetBox(box) 
        self.id_ENERGY = oasysgui.lineEdit(box1, self, "ENERGY",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 2 
        idx += 1 
        box1 = gui.widgetBox(box) 
        self.id_CUR = oasysgui.lineEdit(box1, self, "CUR",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 3 
        idx += 1 
        box1 = gui.widgetBox(box) 
        self.id_PERIOD = oasysgui.lineEdit(box1, self, "PERIOD",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 4 
        idx += 1 
        box1 = gui.widgetBox(box) 
        self.id_N = oasysgui.lineEdit(box1, self, "N",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 

        # COMMENTED AS IT IS NOT IMPLEMENTED!!
        #widget index 5 
        # idx += 1
        # box1 = gui.widgetBox(box)
        # oasysgui.lineEdit(box1, self, "KX",
        #              label=self.unitLabels()[idx], addSpace=False,
        #             valueType=float, orientation="horizontal", labelWidth=250)
        # self.show_at(self.unitFlags()[idx], box1)
        
        #widget index 6 
        idx += 1 
        box1 = gui.widgetBox(box) 
        self.id_KY = oasysgui.lineEdit(box1, self, "KY",
                     valueType=float, label=self.unitLabels()[idx], addSpace=False, orientation="horizontal", labelWidth=250, callback=self.calculateGAP)
        self.show_at(self.unitFlags()[idx], box1)

        #widget index 7
        idx += 1
        box1 = gui.widgetBox(box)
        self.id_GAP = oasysgui.lineEdit(box1, self, "GAP",
                     valueType=float, label=self.unitLabels()[idx], addSpace=False, orientation="horizontal", labelWidth=250, callback=self.calculateKY)
        self.show_at(self.unitFlags()[idx], box1)

        #widget index 8
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "EMIN",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 9
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "EMAX",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 10
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "NEE",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=int, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 11
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "D",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 12
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "XPC",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 13
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "YPC",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 14
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "XPS",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 15
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "YPS",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 16
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "NXP",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=int, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1) 
        
        #widget index 17
        idx += 1 
        box1 = gui.widgetBox(box) 
        oasysgui.lineEdit(box1, self, "NYP",
                     label=self.unitLabels()[idx], addSpace=False,
                    valueType=int, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1)

        self.calculateKY()

    def calculateGAP(self):
        self.GAP=numpy.round((-1 / 0.02677) * numpy.log(self.KY / (93.4 * 2.89408 * self.PERIOD * 0.01)), 5)

    def calculateKY(self):
        self.KY=numpy.round((93.4 * 0.01 * self.PERIOD * (2.3333 * numpy.exp(-0.02473 * self.GAP) + 1.189 * numpy.exp(-0.059691 * self.GAP))), 5)


    def unitLabels(self):
         # return ['Beam energy (GeV)','Beam current (mA)','Period (cm)','Number of periods','Kx','Ky','GAP (mm)','Min energy (eV)','Max energy (eV)','Number of energy steps','Distance (m)','X-pos. (mm)','Y-pos. (mm)','X slit [mm or mrad]','Y slit [mm or mrad]','Integration points X','Integration points Y']
         return ['Beam energy (GeV)','Beam current (mA)','Period (cm)','Number of periods','Ky','GAP (mm)','Min energy (eV)','Max energy (eV)','Number of energy steps','Distance (m)','X-pos. (mm)','Y-pos. (mm)','X slit [mm or mrad]','Y slit [mm or mrad]','Integration points X','Integration points Y']

    def unitFlags(self):
         return ['True','True','True','True','True','True','True','True','True','True','True','True','True','True','True','True']

    def get_help_name(self):
        return 'ID19ws'

    def check_fields(self):


        self.ENERGY = congruence.checkStrictlyPositiveNumber(self.ENERGY, "Beam Energy")
        self.CUR = congruence.checkStrictlyPositiveNumber(self.CUR, "Beam Current")
        self.PERIOD = congruence.checkStrictlyPositiveNumber(self.PERIOD, "Period")
        self.N = congruence.checkStrictlyPositiveNumber(self.N, "Number of Periods")
        self.KX = congruence.checkNumber(self.KX, "Kx")
        self.KY = congruence.checkPositiveNumber(self.KY, "Ky")
        self.GAP = congruence.checkPositiveNumber(self.GAP, "Gap")
        self.EMIN = congruence.checkPositiveNumber(self.EMIN, "Min Energy")
        self.EMAX = congruence.checkStrictlyPositiveNumber(self.EMAX, "Max Energy")
        congruence.checkLessThan(self.EMIN, self.EMAX, "Min Energy", "Max Energy")
        self.NEE = congruence.checkStrictlyPositiveNumber(self.NEE, "Number of energy steps")
        self.D = congruence.checkPositiveNumber(self.D, "Distance")
        self.XPC = congruence.checkNumber(self.XPC, "X-pos")
        self.YPC = congruence.checkNumber(self.YPC, "Y-pos")
        self.XPS = congruence.checkNumber(self.XPS, "X slit")
        self.YPS = congruence.checkNumber(self.YPS, "Y Slit")
        self.NXP = congruence.checkStrictlyPositiveNumber(self.NXP, "Integration points X")
        self.NYP = congruence.checkStrictlyPositiveNumber(self.NYP, "Integration points Y")


    def do_xoppy_calculation(self):   

        dict_parameters = {
            "ENERGY" : self.ENERGY,
            "CUR"    : self.CUR,
            "PERIOD" : self.PERIOD,
            "N"      : self.N,
            "KX"     : self.KX,
            "KY"     : self.KY,
            "EMIN"   : self.EMIN,
            "EMAX"   : self.EMAX,
            "NEE"    : self.NEE,
            "D"      : self.D,
            "XPC"    : self.XPC,
            "YPC"    : self.YPC,
            "XPS"    : self.XPS,
            "YPS"    : self.YPS,
            "NXP"    : self.NXP,
            "NYP"    : self.NYP,
        }

        script = self.script_template().format_map(dict_parameters)

        self.xoppy_script.set_code(script)

        outFile = xoppy_calc_ws(
            ENERGY = self.ENERGY,
            CUR    = self.CUR,
            PERIOD = self.PERIOD,
            N      = self.N,
            KX     = self.KX,
            KY     = self.KY,
            EMIN   = self.EMIN,
            EMAX   = self.EMAX,
            NEE    = self.NEE,
            D      = self.D,
            XPC    = self.XPC,
            YPC    = self.YPC,
            XPS    = self.XPS,
            YPS    = self.YPS,
            NXP    = self.NXP,
            NYP    = self.NYP,
        )
        return outFile, script

    def script_template(self):
        return """
#
# script to make the calculations (created by XOPPY:WS)
#
import numpy
from xoppylib.xoppy_run_binaries import xoppy_calc_ws
out_file =  xoppy_calc_ws(
        ENERGY = {ENERGY},
        CUR    = {CUR},
        PERIOD = {PERIOD},
        N      = {N},
        KX     = {KX},
        KY     = {KY},
        EMIN   = {EMIN},
        EMAX   = {EMAX},
        NEE    = {NEE},
        D      = {D},
        XPC    = {XPC},
        YPC    = {YPC},
        XPS    = {XPS},
        YPS    = {YPS},
        NXP    = {NXP},
        NYP    = {NYP},
        )
# data to pass to power
data = numpy.loadtxt(out_file)
energy = data[:,0]
flux = data[:,1]
spectral_power = data[:,2]
cumulated_power = data[:,3]
#
# example plot
#
from srxraylib.plot.gol import plot
plot(energy,flux,
    xtitle="Photon energy [eV]",ytitle="Flux [photons/s/0.1%bw]",title="WS Flux",
    xlog=True,ylog=True,show=False)
plot(energy,spectral_power,
    xtitle="Photon energy [eV]",ytitle="Power [W/eV]",title="WS Spectral Power",
    xlog=True,ylog=True,show=False)
plot(energy,cumulated_power,
    xtitle="Photon energy [eV]",ytitle="Cumulated Power [W]",title="WS Cumulated Power",
    xlog=False,ylog=False,show=True)
    
#
# end script
#
"""

    def extract_data_from_xoppy_output(self, calculation_output):
    
            spec_file_name, script = calculation_output
            out = numpy.loadtxt(spec_file_name)
            if len(out) == 0: raise Exception("Calculation gave no results (empty data)")
    
            calculated_data = DataExchangeObject("XOPPY", self.get_data_exchange_widget_name())
            calculated_data.add_content("xoppy_specfile", spec_file_name)
            calculated_data.add_content("xoppy_data", out)
            calculated_data.add_content("xoppy_script", script)
    
            return calculated_data
     

    def get_data_exchange_widget_name(self):
        return "WS"  

    def getTitles(self):
        return ['Flux','Spectral power','Cumulated power']

    def getXTitles(self):
        return ["Energy [eV]","Energy [eV]","Energy [eV]"]

    def getYTitles(self):
        return ["Flux [Photons/sec/0.1%bw]","Spectral Power [W/eV]","Cumulated Power [W]"]

    def getVariablesToPlot(self):
        return [(0, 1),(0, 2),(0, 3)]

    def getLogPlot(self):
        return [(True, True),(True, True),(False, False)]

    def receive_syned_data(self, data):

        if isinstance(data, synedb.Beamline):
            if not data._light_source is None and isinstance(data._light_source._magnetic_structure, synedid.InsertionDevice):
                light_source = data._light_source

                self.N = int(light_source._magnetic_structure._number_of_periods)
                self.ENERGY = light_source._electron_beam._energy_in_GeV
                self.CUR = 1e3*light_source._electron_beam._current
                self.PERIOD = 1e2*light_source._magnetic_structure._period_length
                self.KY = light_source._magnetic_structure._K_vertical
                self.GAP=(-1/0.02677)*numpy.log(self.KY/(93.4*2.89408*self.PERIOD*0.01))
                self.set_enabled(False)
            else:
                self.set_enabled(True)
        else:
            self.set_enabled(True)

    def set_enabled(self,value):
        if value == True:
                self.id_N.setEnabled(True)
                self.id_ENERGY.setEnabled(True)
                self.id_CUR.setEnabled(True)
                self.id_PERIOD.setEnabled(True)
                self.id_KY.setEnabled(True)
                self.id_GAP.setEnabled(True)
        else:
                self.id_N.setEnabled(False)
                self.id_ENERGY.setEnabled(False)
                self.id_CUR.setEnabled(False)
                self.id_PERIOD.setEnabled(False)
                self.id_KY.setEnabled(False)
                self.id_GAP.setEnabled(False)


# --------------------------------------------------------------------------------------------




if __name__ == "__main__":
    import os, sys
    os.environ['LD_LIBRARY_PATH'] = ''
    app = QApplication(sys.argv)
    w = OWID19ws()
    w.show()
    app.exec()
    w.saveSettings()




from syned.beamline.beamline import Beamline
from syned.storage_ring.light_source import LightSource
from syned.storage_ring.empty_light_source import EmptyLightSource

class WOBeamline(Beamline):

    def __init__(self,
                 light_source=None,
                 beamline_elements_list=None):
        super().__init__(light_source=light_source, beamline_elements_list=beamline_elements_list)

        self._propagation_info_list = [{}] * super().get_beamline_elements_number()


    def duplicate(self):
        beamline_elements_list = []
        for beamline_element in self._beamline_elements_list:
            beamline_elements_list.append(beamline_element)

        return WOBeamline(light_source=self._light_source,
                        beamline_elements_list=beamline_elements_list)

    def append_beamline_element(self, beamline_element, propagation_info={}):
        super().append_beamline_element(beamline_element)
        self._propagation_info_list.append(propagation_info)

    def get_propagation_info_list(self):
        return self._propagation_info_list

    def get_propagation_info_at(self, i):
        return self.get_propagation_info_list()[i]

    def to_python_code(self,do_plot=True):

        text_code = "\nimport numpy"
        text_code += "\n\n"

        text_code += "\n#"
        text_code += "\n# Import section (propagators)"
        text_code += "\n#"
        text_code += "\nfrom wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters"
        text_code += "\nfrom syned.beamline.beamline_element import BeamlineElement"
        text_code += "\nfrom syned.beamline.element_coordinates import ElementCoordinates"
        text_code += "\nfrom wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"
        text_code += "\nfrom wofry.propagator.propagators2D.fresnel import Fresnel2D"
        text_code += "\nfrom wofry.propagator.propagators2D.fresnel_convolution import FresnelConvolution2D"
        text_code += "\nfrom wofry.propagator.propagators2D.fraunhofer import Fraunhofer2D"
        text_code += "\nfrom wofry.propagator.propagators2D.integral import Integral2D"
        text_code += "\nfrom wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"


        text_code  +=  "\n\n\n##########  SOURCE ##########\n\n\n"
        text_code += self.get_light_source().to_python_code(do_plot=do_plot)

        if self.get_beamline_elements_number() > 0:
            text_code += "\n\n\n##########  OPTICAL SYSTEM ##########\n\n\n"


            for index in range(self.get_beamline_elements_number()):
                text_code += "\n\n\n##########  OPTICAL ELEMENT NUMBER %i ##########\n\n\n" % (index+1)
                oe_name = "oe_" + str(index)
                beamline_element = self.get_beamline_element_at(index)
                optical_element = beamline_element.get_optical_element()
                coordinates = beamline_element.get_coordinates()

                text_code += "\ninput_wavefront = output_wavefront.duplicate()"

                # OPTICAL ELEMENT ----------------
                text_code += optical_element.to_python_code(do_plot=do_plot)

                if (coordinates.p() == 0.0) and (coordinates.q() == 0.0): # NO DRIFT
                    text_code += "\n# no drift in this element "
                    text_code += "\noutput_wavefront = optical_element.applyOpticalElement(input_wavefront)"
                else:
                    if coordinates.p() != 0.0:
                        text_code += "\n# drift_before %g m" % coordinates.p()
                    if coordinates.q() != 0.0:
                        text_code += "\n# drift_after %g m " % coordinates.q()

##########################
# 1D
# ==
#
# propagators_list = ["Fresnel",    "Fresnel (Convolution)",  "Fraunhofer",    "Integral",    "Fresnel Zoom",    "Fresnel Zoom Scaled"]
# class_name       = ["Fresnel1D",  "FresnelConvolution1D",   "Fraunhofer1D",  "Integral1D",  "FresnelZoom1D",   "FresnelZoomScaling1D"]
# handler_name     = ["FRESNEL_1D", "FRESNEL_CONVOLUTION_1D", "FRAUNHOFER_1D", "INTEGRAL_1D", "FRESNEL_ZOOM_1D", "FRESNEL_ZOOM_SCALING_1D"]
#
# 2D
# ==
# propagators_list = ["Fresnel",   "Fresnel (Convolution)",  "Fraunhofer",    "Integral",    "Fresnel Zoom XY"   ]
# class_name       = ["Fresnel2D", "FresnelConvolution2D",   "Fraunhofer2D",  "Integral2D",  "FresnelZoomXY2D"   ]
# handler_name     = ["FRESNEL_2D","FRESNEL_CONVOLUTION_2D", "FRAUNHOFER_2D", "INTEGRAL_2D", "FRESNEL_ZOOM_XY_2D"]

                    propagation_info = self.get_propagation_info_at(index)

                    propagator_class_name                   = propagation_info["propagator_class_name"]
                    propagator_handler_name                 = propagation_info["propagator_handler_name"]
                    propagator_additional_parameters_names  = propagation_info["propagator_additional_parameters_names"]
                    propagator_additional_parameters_values = propagation_info["propagator_additional_parameters_values"]


                    # text_code += "\n#"
                    # text_code += "\n# Import section"
                    # text_code += "\n#"
                    # text_code += "\nimport numpy"
                    # text_code += "\nfrom wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters"
                    # text_code += "\nfrom syned.beamline.beamline_element import BeamlineElement"
                    # text_code += "\nfrom syned.beamline.element_coordinates import ElementCoordinates"
                    # text_code += "\nfrom wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"
                    # text_code += "\nfrom wofry.propagator.propagators2D.fresnel import Fresnel2D"
                    # text_code += "\nfrom wofry.propagator.propagators2D.fresnel_convolution import FresnelConvolution2D"
                    # text_code += "\nfrom wofry.propagator.propagators2D.fraunhofer import Fraunhofer2D"
                    # text_code += "\nfrom wofry.propagator.propagators2D.integral import Integral2D"
                    # text_code += "\nfrom wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"
                    text_code += "\n#"
                    text_code += "\n# propagating\n#"
                    text_code += "\n#"
                    text_code += "\npropagation_elements = PropagationElements()"
                    text_code += "\nbeamline_element = BeamlineElement(optical_element=optical_element,"
                    text_code += "    coordinates=ElementCoordinates(p=%f," % (coordinates.p())
                    text_code += "    q=%f," % (coordinates.q())
                    text_code += "    angle_radial=numpy.radians(%f)," % (coordinates.angle_radial())
                    text_code += "    angle_azimuthal=numpy.radians(%f)))" % (coordinates.angle_azimuthal())
                    text_code += "\npropagation_elements.add_beamline_element(beamline_element)"
                    text_code += "\npropagation_parameters = PropagationParameters(wavefront=input_wavefront,"
                    text_code += "    propagation_elements = propagation_elements)"
                    text_code += "\n#self.set_additional_parameters(propagation_parameters)"
                    text_code += "\n#"

                    for i in range(len(propagator_additional_parameters_names)):
                        text_code += "\npropagation_parameters.set_additional_parameters('%s', %s)" % \
                        (propagator_additional_parameters_names[i], str(propagator_additional_parameters_values[i]))

                    text_code += "\n#"
                    text_code += "\npropagator = PropagationManager.Instance()"
                    text_code += "\ntry:"
                    text_code += "\n    propagator.add_propagator(%s())" % propagator_class_name
                    text_code += "\nexcept:"
                    text_code += "\n    pass"
                    text_code += "\noutput_wavefront = propagator.do_propagation(propagation_parameters=propagation_parameters,"
                    text_code += "    handler_name='%s')" % (propagator_handler_name)
##########################

                if do_plot:
                        text_code += "\n\n\nfrom srxraylib.plot.gol import plot, plot_image"
                        text_code += "\nif output_wavefront.get_dimension() == 1:"
                        text_code += "\n  plot(output_wavefront.get_abscissas(),output_wavefront.get_intensity(),title='OPTICAL ELEMENT NR %d')" % (index+1)
                        text_code += "\nelse:"
                        text_code += "\n  plot_image(output_wavefront.get_intensity(),output_wavefront.get_coordinate_x(),output_wavefront.get_coordinate_y(),aspect='auto',title='OPTICAL ELEMENT NR %d')" % (index+1)

        #######################
        # txt = "#"
        # txt += "\n# Import section"
        # txt += "\n#"
        # txt += "\nimport numpy"
        # txt += "\nfrom wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters"
        # txt += "\nfrom syned.beamline.beamline_element import BeamlineElement"
        # txt += "\nfrom syned.beamline.element_coordinates import ElementCoordinates"
        # txt += "\nfrom wofry.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"
        #
        # if write_wavefront_template:
        #     txt += "\n\n#"
        #     txt += "\n# create/import your input_wavefront (THIS IS A PLACEHOLDER - REPLACE WITH YOUR SOURCE)"
        #     txt += "\n#"
        #     txt += "\nfrom wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D"
        #     txt += "\ninput_wavefront = GenericWavefront2D.load_h5_file('wavefront2D_input.h5',filepath='wfr')"
        #     txt += "\n\n"
        #
        # txt += "\n\n#"
        # txt += "\n# info on current oe\n#"
        # txt += "\n#"
        # txt_info = self.get_optical_element().info()
        # lines = txt_info.split("\n")
        #
        # for line in lines:
        #     txt += "\n#"+line
        #
        # txt += "\n\n#"
        # txt += "\n# define current oe"
        # txt += "\n#"
        #
        # txt += self.get_optical_element_python_code()
        #
        # txt += "\n#"
        # txt += "\n# propagating (***  ONLY THE ZOOM PROPAGATOR IS IMPLEMENTED ***)\n#"
        # txt += "\n#"
        #
        # txt += "\npropagation_elements = PropagationElements()"
        # txt += "\nbeamline_element = BeamlineElement(optical_element=optical_element,"
        # txt += "    coordinates=ElementCoordinates(p=%f,"%(self.p)
        # txt += "    q=%f,"%(self.q)
        # txt += "    angle_radial=numpy.radians(%f),"%(self.angle_radial)
        # txt += "    angle_azimuthal=numpy.radians(%f)))"%(self.angle_azimuthal)
        # txt += "\npropagation_elements.add_beamline_element(beamline_element)"
        # txt += "\npropagation_parameters = PropagationParameters(wavefront=input_wavefront.duplicate(),"
        # txt += "    propagation_elements = propagation_elements)"
        # txt += "\n#self.set_additional_parameters(propagation_parameters)"
        #
        # txt += "\n#"
        # txt += "\npropagation_parameters.set_additional_parameters('shift_half_pixel', 1)"
        # txt += "\npropagation_parameters.set_additional_parameters('magnification_x', %f)"%(self.magnification_x)
        # txt += "\npropagation_parameters.set_additional_parameters('magnification_y', %f)"%(self.magnification_y)
        #
        # txt += "\n#"
        # txt += "\npropagator = PropagationManager.Instance()"
        # txt += "\ntry:"
        # txt += "\n    propagator.add_propagator(FresnelZoomXY2D())"
        # txt += "\nexcept:"
        # txt += "\n    pass"
        # txt += "\noutput_wavefront = propagator.do_propagation(propagation_parameters=propagation_parameters,"
        # txt += "    handler_name='FRESNEL_ZOOM_XY_2D')"
        #
        # txt += "\n\nfrom srxraylib.plot.gol import plot_image"
        # txt += "\nplot_image(output_wavefront.get_intensity(),output_wavefront.get_coordinate_x(), output_wavefront.get_coordinate_y(), aspect='auto')"

#######################
        return text_code


import numpy
from oasys.util.oasys_util import get_fwhm

from srxraylib.plot.gol import plot, plot_image
import matplotlib.pylab as plt
import os

# def get_fwhm(histogram, bins):
#     quote = numpy.max(histogram)*0.5
#     cursor = numpy.where(histogram >= quote)
#
#     if histogram[cursor].size > 1:
#         bin_size    = bins[1]-bins[0]
#         fwhm        = bin_size*(cursor[0][-1]-cursor[0][0])
#         coordinates = (bins[cursor[0][0]], bins[cursor[0][-1]])
#     else:
#         fwhm = 0.0
#         coordinates = None
#
#     return fwhm, quote, coordinates

#
#
#
class Tally():
    def __init__(self,
                 scan_variable_name='x',
                 additional_stored_variable_names=None,
                 do_store_wavefronts=False):
        self.reset()
        self.scan_variable_name = scan_variable_name
        self.additional_stored_variable_names = additional_stored_variable_names
        self.do_store_wavefronts = do_store_wavefronts

    def reset(self):
        self.scan_variable_index = -1
        self.scan_variable_value = []
        self.fwhm = []
        self.intensity_at_center = []
        self.intensity_total = []
        self.intensity_peak = []
        self.additional_stored_values = []
        self.stored_wavefronts = []


    def append(self, wf, scan_variable_value=None, additional_stored_values=None):
        fwhm, intensity_total, intensity_at_center, intensity_peak = self.process_wavefront(wf)
        self.fwhm.append(fwhm)
        self.intensity_at_center.append(intensity_at_center)
        self.intensity_total.append(intensity_total)
        self.intensity_peak.append(intensity_peak)
        self.scan_variable_index += 1
        if scan_variable_value is None:
            self.scan_variable_value.append(self.scan_variable_index)
        else:
            self.scan_variable_value.append(scan_variable_value)

        self.additional_stored_values.append(additional_stored_values)

        if self.do_store_wavefronts:
            self.stored_wavefronts.append(wf.duplicate())

    def get_wavefronts(self):
        return self.stored_wavefronts

    def get_number_of_calls(self):
        return self.scan_variable_index + 1

    def get_additional_stored_values(self):
        return self.additional_stored_values

    def get_scan_variable_value(self):
        return numpy.array(self.scan_variable_value)

    def get_intensity_at_center(self):
        return numpy.array(self.intensity_at_center)

    def get_fwhm(self):
        return numpy.array(self.fwhm)

    def get_wavefronts_intensity(self):
        if len(self.stored_wavefronts) == 0:
            raise Exception("No stored wavefronts found")

        for i, wf in enumerate(self.stored_wavefronts):
            x, y = wf.get_abscissas(), wf.get_intensity()
            if i == 0:
                INTENSITY = numpy.zeros((self.get_number_of_calls(), x.size))
            INTENSITY[i, :] = y
        return INTENSITY

    def get_wavefronts_abscissas(self):
        if len(self.stored_wavefronts) == 0:
            raise Exception("No stored wavefronts found")
        else:
            return self.stored_wavefronts[-1].get_abscissas()

    def save_scan(self, filename="tmp.dat", add_header=True):
        f = open(filename, 'w')
        if add_header:
            if self.additional_stored_variable_names is None:
                number_of_additional_parameters = 0
            else:
                number_of_additional_parameters = len(self.additional_stored_variable_names)
            header = "#S 1 scored data\n"
            header += "#N %d\n" % (number_of_additional_parameters + 5)
            header_titles = "#L  %s  %s  %s  %s  %s" % (self.scan_variable_name, "fwhm", "total_intensity", "on_axis_intensity", "peak_intensity")
            for i in range(number_of_additional_parameters):
                header_titles += "  %s" % self.additional_stored_variable_names[i]
            header_titles += "\n"
            header += header_titles
            f.write(header)
        for i in range(len(self.fwhm)):
            f.write("%g %g %g %g %g" % (self.scan_variable_value[i],
                                    1e6*self.fwhm[i],
                                    self.intensity_total[i],
                                    self.intensity_at_center[i],
                                    self.intensity_peak[i]))
            for j in range(number_of_additional_parameters):
                f.write(" %g" % self.additional_stored_values[i][j])
            f.write("\n")
        f.close()
        print("File written to disk: %s" % filename)

    def plot(self, title="", factor_abscissas=1.0, xtitle=None):
        self.plot_fwhm(title=title, factor_abscissas=factor_abscissas, xtitle=xtitle)
        self.plot_intensity_at_center(title=title, factor_abscissas=factor_abscissas, xtitle=xtitle)

    def plot_intensity_at_center(self, title="", factor_abscissas=1.0, xtitle=None):
        x = numpy.array(self.scan_variable_value)
        y = numpy.array(self.intensity_at_center)
        if xtitle is None:
            xtitle = self.scan_variable_name
        out = plot(factor_abscissas * x, y, yrange=[0,1.1*y.max()],
             title=title, ytitle="Intensity at center[a.u.]", xtitle=xtitle,
             figsize=(15, 4), show=0)
        return out

    def plot_fwhm(self, title="", factor_fwhm=1.0, xtitle=None, ytitle=None):
        x = numpy.array(self.scan_variable_value)
        y = numpy.array(self.fwhm)
        if xtitle is None:
            xtitle = self.scan_variable_name
        if ytitle is None:
            if factor_fwhm == 1.0:
                ytitle = "FWHM [m]"
            elif factor_fwhm == 1e6:
                ytitle = "FWHM [um]"
            else:
                ytitle = "FWHM"

        out = plot(x, factor_fwhm * y, yrange=[0,1.1*factor_fwhm*y.max()],
             title=title, ytitle=ytitle, xtitle=xtitle,
             figsize=(15, 4), show=1)
        return out

    def plot_wavefronts_intensity(self,
                                  xtitle="scan_variable_value",
                                  ytitle="wavefront abscissas",
                                  factor_abscissas=1.0,
                                  title=""):

        out = plot_image(self.get_wavefronts_intensity(),
                   self.get_scan_variable_value(),
                   factor_abscissas * self.get_wavefronts_abscissas(),
                   xtitle=xtitle, ytitle=ytitle, title=title, aspect='auto')
        return out

    @classmethod
    def process_wavefront(cls, wf):
        I = wf.get_intensity()
        x = wf.get_abscissas()

        fwhm, quote, coordinates = get_fwhm(I, x)
        intensity_at_center = I[I.size // 2]
        intensity_total = I.sum() * (x[1] - x[0])
        intensity_peak = I.max()

        return fwhm, intensity_total, intensity_at_center, intensity_peak



class TallyCoherentModes(Tally):
    def __init__(self,
                 additional_stored_variable_names=None):

        super().__init__(scan_variable_name='mode_index',
                 additional_stored_variable_names=additional_stored_variable_names,
                 do_store_wavefronts=True)

        self.abscissas = None
        self.cross_spectral_density = None
        self.spectral_density = None,
        self.eigenvalues = None
        self.eigenvectors = None

    def get_cross_pectral_density(self):
        if self.cross_spectral_density is None: self.calculate_cross_spectral_density()
        return self.cross_spectral_density

    def get_spectral_density_from_intensities(self):
        WF = self.get_wavefronts()
        print(WF)
        intensity = None
        for i,wf in enumerate(WF):
            if intensity is None:
                intensity = wf.get_intensity()
            else:
                intensity += wf.get_intensity()
        return intensity


    def get_spectral_density(self):
        csd = self.get_cross_pectral_density()
        nx = csd.shape[0]
        spectral_density = numpy.zeros(nx)
        for i in range(nx):
            spectral_density[i] = csd[i, i]
        return spectral_density

    def get_eigenvalues(self):
        if self.eigenvalues is None: self.diagonalize()
        return self.eigenvalues

    def get_eigenvectors(self):
        if self.eigenvectors is None: self.diagonalize()
        return self.eigenvectors

    def get_abscissas(self):
        if self.abscissas is None: self.abscissas = self.get_wavefronts()[-1].get_abscissas()
        return self.abscissas


    def calculate_cross_spectral_density(self, do_plot=False):
        # retrieve arrays
        WFs = self.get_wavefronts()
        nmodes = self.get_number_of_calls()
        abscissas = self.get_abscissas()
        #
        # calculate the CSD
        #

        input_array = numpy.zeros((nmodes, abscissas.size), dtype=complex)
        for i,wf in enumerate(WFs):
            input_array[i,:] = wf.get_complex_amplitude() # tmp[i][0]

        cross_spectral_density = numpy.zeros((abscissas.size, abscissas.size), dtype=complex)

        for i in range(nmodes):
            cross_spectral_density += numpy.outer(numpy.conjugate(input_array[i, :]), input_array[i, :])

        self.cross_spectral_density = cross_spectral_density


    def diagonalize(self, do_plot=False):
        csd = self.get_cross_pectral_density()

        #
        # diagonalize the CSD
        #

        w, v = numpy.linalg.eig(csd)
        print(w.shape, v.shape)
        idx = w.argsort()[::-1]  # large to small
        self.eigenvalues = numpy.real(w[idx])
        self.eigenvectors = v[:, idx].T


    def get_occupation(self):
        ev = self.get_eigenvalues()
        return  numpy.arange(ev.size), ev / ev.sum()


    def calculate_coherent_fraction(self, do_plot=False):
        if self.eigenvalues is None:
            self.diagonalize()
        cf = self.eigenvalues[0] / self.eigenvalues.sum()
        return cf, self.eigenvalues, self.eigenvectors, self.cross_spectral_density

    def plot_cross_spectral_density(self, show=True, filename=""):
        csd = self.get_cross_pectral_density()
        plot_image(numpy.abs(csd), 1e6*self.abscissas, 1e6*self.abscissas,
                   title="Cross Spectral Density", xtitle="X1 [um]", ytitle="X2 [um]",show=False)

        if filename != "":
            plt.savefig(filename)
            print("File written to disk: %s" % filename)

        if show:
            plt.show()
        else:
            plt.close()

        print("matrix cross_spectral_density: ", csd.shape)

    def plot_spectral_density(self, show=True, filename="", method=2, title=""):
        #
        # plot intensity
        #
        abscissas = self.get_abscissas()
        eigenvalues = self.get_eigenvalues()
        eigenvectors = self.get_eigenvectors()

        spectral_density = self.get_spectral_density() # numpy.zeros_like(abscissas)
        fwhm, quote, coordinates = get_fwhm(spectral_density, 1e6 * abscissas)

        if method > 0:
            nmodes = self.get_number_of_calls()
            y = numpy.zeros_like(abscissas)
            for i in range(nmodes):
                y += eigenvalues[i] * numpy.real(numpy.conjugate(eigenvectors[i, :]) * eigenvectors[i, :])

        if method == 0:
            plot(1e6 * abscissas, spectral_density,
                 xtitle="x [um]", ytitle="Spectral Density (From Cross Spectral Density)",
                 title="%s FWHM = %g um" % (title, fwhm), show=False)
        elif method == 1:
            plot(1e6 * abscissas, y,
                 xtitle="x [um]", ytitle="Spectral Density (From modes)", title="%s FWHM = %g um" % (title,fwhm), show=False)
        elif method == 2:
            plot(1e6 * abscissas, spectral_density,
                 1e6 * abscissas, y, legend=["From Cross Spectral Density", "From modes"],
                 xtitle="x [um]", ytitle="Spectral Density", title="%s FWHM = %g um" % (title, fwhm), show=False)


        if filename != "":
            plt.savefig(filename)
            print("File written to disk: %s" % filename)

        if show:
            plt.show()
        else:
            plt.close()

    def save_spectral_density(self, filename="", add_header=True):
        #
        # plot intensity
        #
        abscissas = self.get_abscissas()

        spectral_density = self.get_spectral_density()
        fwhm, quote, coordinates = get_fwhm(spectral_density, 1e6 * abscissas)


        f = open(filename, 'w')
        if add_header:
            header = "#S 1 spectral density\n#N 2\n#UFWHM %g\n" % fwhm
            header += "#L  %s  %s\n" % ("abscissas [um]","spectral density")
            f.write(header)

        for i in range(abscissas.size):
            f.write("%g  %g\n" % (1e6 * abscissas[i], spectral_density[i]))
        f.close()
        print("File written to disk: %s" % filename)


    def plot_occupation(self, show=True, filename=""):
        x, y = self.get_occupation()
        nmodes = self.get_number_of_calls()
        plot(x[0:nmodes], y[0:nmodes],
             title="CF: %g" % (y[0]),
             xtitle="mode index", ytitle="occupation", show=False)

        if filename != "":
            plt.savefig(filename)
            print("File written to disk: %s" % filename)

        if show:
            plt.show()
        else:
            plt.close()

    def save_occupation(self, filename="", add_header=True):
        x, y = self.get_occupation()
        nmodes = self.get_number_of_calls()


        f = open(filename, 'w')
        if add_header:
            header = "#S 1 occupation\n#N 2\n"
            header += "#L  %s  %s\n" % ("mode index","occupation")
            f.write(header)

        for i in range(nmodes):
            f.write("%g  %g\n" % (x[i], y[i]))
        f.close()
        print("File written to disk: %s" % filename)


if __name__ == "__main__":
    from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D

    # sc = Tally(scan_variable_name='mode index', additional_stored_variable_names=['a', 'b'])
    sc = TallyCoherentModes()
    for xmode in range(51):
        output_wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.00012, x_max=0.00012,
                                                                              number_of_points=1000)
        output_wavefront.set_photon_energy(10000)
        output_wavefront.set_gaussian_hermite_mode(sigma_x=3.03783e-05, amplitude=1, mode_x=xmode, shift=0, beta=0.0922395)


        sc.append(output_wavefront, scan_variable_value=xmode, additional_stored_values=[1,2.1])

    # sc.plot()
    sc.save_scan("tmp.dat")

    # plot(sc.get_abscissas(), sc.get_spectral_density_from_intensities(), title="Spectral Density from intensities")

    sc.plot_cross_spectral_density(show=1,filename="tmp_cs.png")
    sc.plot_spectral_density(show=1, filename="tmp_sd.png", method=2)
    sc.save_spectral_density(filename="tmp_sd.txt")
    sc.plot_occupation(show=1, filename="tmp_occ.png",)
    sc.save_occupation(filename="tmp_occ.txt")


    # cf, _, _, _ = sc.calculate_coherent_fraction(do_plot=1)


import sys
from PyQt5 import QtGui, QtWidgets
from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtWidgets import QApplication, QFileDialog
from PyQt5.QtCore import QRect
from PyQt5.QtGui import QTextCursor

from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.widgets import widget
from oasys.util.oasys_util import TriggerIn, TriggerOut, EmittingStream

from orangecontrib.wofry.util.wofry_objects import WofryData
from orangecontrib.esrf.shadow.util.python_script import PythonConsole


# from oasys.widgets import gui as oasysgui
# from oasys.widgets import congruence
# from oasys.widgets import congruence
# from oasys.util.oasys_util import TriggerIn, TriggerOut, EmittingStream
# from syned.storage_ring.magnetic_structures.undulator import Undulator
# from syned.beamline.beamline import Beamline
# import inspect
# import numpy
# import os, sys

from wofryimpl.propagator.light_source import WOLightSource
from orangecontrib.esrf.wofry.util.light_source import WOLightSourceCMD


class DiagonalizePythonScript(widget.OWWidget):

    name = "Diagonalize Python Script"
    description = "Diagonalize Python Script"
    icon = "icons/diagonalize_python_script.png"
    maintainer = "Manuel Sanchez del Rio"
    maintainer_email = "srio(@at@)esrf.eu"
    priority = 500
    category = "Tools"
    keywords = ["script"]

    inputs = [("WofryData", WofryData, "set_input")]

    outputs = [
               {"name":"WofryData",
                "type":WofryData,
                "doc":"WofryData",
                "id":"WofryData"}
                ]

    mode_index_max = Setting(10)

    show_graph_flag = Setting(1)
    script_file_flag = Setting(0)
    script_file_name = Setting("tmp.py")
    graph_file_flag = Setting(0)
    data_file_flag = Setting(0)
    root_file_name = Setting("tmp")

    #
    #
    #
    IMAGE_WIDTH = 890
    IMAGE_HEIGHT = 680

    # want_main_area=1

    is_automatic_run = Setting(True)

    error_id = 0
    warning_id = 0
    info_id = 0

    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 560

    input_data = None



    def __init__(self, show_automatic_box=True, show_general_option_box=True):
        super().__init__() # show_automatic_box=show_automatic_box)


        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, self.MAX_WIDTH)),
                               round(min(geom.height()*0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        self.general_options_box = gui.widgetBox(self.controlArea, "General Options", addSpace=True, orientation="horizontal")
        self.general_options_box.setVisible(show_general_option_box)

        if show_automatic_box :
            gui.checkBox(self.general_options_box, self, 'is_automatic_run', 'Automatic Execution')


        #
        #
        #
        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Refresh Script", callback=self.refresh_script)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)


        gui.separator(self.controlArea)

        gen_box = oasysgui.widgetBox(self.controlArea, "Scan modes", addSpace=False, orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)
        oasysgui.lineEdit(gen_box, self, "mode_index_max", "Number of modes", labelWidth=150, valueType=int,
                          orientation="horizontal", callback=self.refresh_script)

        gen_box = oasysgui.widgetBox(self.controlArea, "Script file", addSpace=False, orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)
        gui.comboBox(gen_box, self, "script_file_flag", label="write file with script",
                     items=["No", "Yes"], labelWidth=300,
                     sendSelectedValue=False, orientation="horizontal")
        box1 = gui.widgetBox(gen_box, orientation="horizontal")
        oasysgui.lineEdit(box1, self, "script_file_name", "Script File Name", labelWidth=150, valueType=str,
                          orientation="horizontal")
        self.show_at("self.script_file_flag == 1", box1)


        gen_box = oasysgui.widgetBox(self.controlArea, "Output Graph and Data", addSpace=False, orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)
        gui.comboBox(gen_box, self, "show_graph_flag", label="show plots",
                     items=["No", "Yes"], labelWidth=300,
                     sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(gen_box, self, "graph_file_flag", label="dump plots to file",
                     items=["No", "Yes [png]", "Yes [pdf]"], labelWidth=300,
                     sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(gen_box, self, "data_file_flag", label="dump plot data to file",
                     items=["No", "Yes"], labelWidth=300,
                     sendSelectedValue=False, orientation="horizontal")

        box3 = gui.widgetBox(gen_box, orientation="horizontal")
        oasysgui.lineEdit(box3, self, "root_file_name", "Data File Name Root", labelWidth=150, valueType=str,
                          orientation="horizontal")

        self.show_at("self.graph_file_flag > 0 or self.data_file_flag == 1", box3)


        tabs_setting = oasysgui.tabWidget(self.mainArea)
        tabs_setting.setFixedHeight(self.IMAGE_HEIGHT)
        tabs_setting.setFixedWidth(self.IMAGE_WIDTH)

        tab_scr = oasysgui.createTabPage(tabs_setting, "Python Script")
        tab_out = oasysgui.createTabPage(tabs_setting, "System Output")

        self.pythonScript = oasysgui.textArea(readOnly=False)
        self.pythonScript.setStyleSheet("background-color: white; font-family: Courier, monospace;")
        self.pythonScript.setMaximumHeight(self.IMAGE_HEIGHT - 250)

        script_box = oasysgui.widgetBox(tab_scr, "", addSpace=False, orientation="vertical", height=self.IMAGE_HEIGHT - 10, width=self.IMAGE_WIDTH - 10)
        script_box.layout().addWidget(self.pythonScript)

        console_box = oasysgui.widgetBox(script_box, "", addSpace=True, orientation="vertical",
                                          height=150, width=self.IMAGE_WIDTH - 10)

        self.console = PythonConsole(self.__dict__, self)
        console_box.layout().addWidget(self.console)

        self.wofry_output = oasysgui.textArea()

        out_box = oasysgui.widgetBox(tab_out, "System Output", addSpace=True, orientation="horizontal", height=self.IMAGE_WIDTH - 45)
        out_box.layout().addWidget(self.wofry_output)

        #############################

        button_box = oasysgui.widgetBox(tab_scr, "", addSpace=True, orientation="horizontal")

        gui.button(button_box, self, "Run Script", callback=self.execute_script, height=40)
        # gui.button(button_box, self, "Save Script to File", callback=self.save_script, height=40)

        gui.rubber(self.controlArea)

        self.process_showers()

    def set_input(self, wofry_data):
        if not wofry_data is None:
            if isinstance(wofry_data, WofryData):
                self.input_data = wofry_data
            else:
                raise Exception("Bad input.")

            if self.is_automatic_run:
                self.refresh_script()


    def callResetSettings(self):
        pass

    def execute_script(self):

        self._script = str(self.pythonScript.toPlainText())
        self.console.write("\nRunning script:\n")
        self.console.push("exec(_script)")
        self.console.new_prompt(sys.ps1)


    def save_script(self):
        # file_name = QFileDialog.getSaveFileName(self, "Save File to Disk", os.getcwd(), filter='*.py')[0]
        file_name = self.script_file_name
        if not file_name is None:
            if not file_name.strip() == "":
                file = open(file_name, "w")
                file.write(str(self.pythonScript.toPlainText()))
                file.close()


    def refresh_script(self):

        self.wofry_output.setText("")

        sys.stdout = EmittingStream(textWritten=self.writeStdOut)

        if self.input_data is None:
            raise Exception("No input data")

        beamline = self.input_data.get_beamline()
        self.pythonScript.setText(to_python_code(beamline,
                                                 do_plot=False,
                                                 mode_index_max=self.mode_index_max,
                                                 show_graph_flag=self.show_graph_flag,
                                                 graph_file_flag=self.graph_file_flag,
                                                 data_file_flag=self.data_file_flag,
                                                 root_file_name=self.root_file_name,
                                                 ))

        if self.script_file_flag:
            self.save_script()

    def writeStdOut(self, text):
        cursor = self.wofry_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.wofry_output.setTextCursor(cursor)
        self.wofry_output.ensureCursorVisible()

def to_python_code(self, # self is beamline
                    do_plot=True,
                    mode_index_max=2,
                    show_graph_flag=1,
                    graph_file_flag=0,
                    data_file_flag=0,
                    root_file_name="tmp",
                   ):

    import_text_code = ""
    import_text_code += "\n#"
    import_text_code += "\n# Import section"
    import_text_code += "\n#"
    import_text_code += "\nimport numpy"
    import_text_code += "\n\nfrom syned.beamline.beamline_element import BeamlineElement"
    import_text_code += "\nfrom syned.beamline.element_coordinates import ElementCoordinates"
    import_text_code += "\nfrom wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters"

    if self.get_light_source() is None:
        source_dimension = 0
        import_text_code += "\n\n#\n# UNDEFINED SOURCE (please complete...)!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! \n#"
        import_text_code += "\noutput_wavefront = None"
        import_text_code += "\n#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
    else:
        source_dimension = self.get_light_source().get_dimension()
        if source_dimension == 1:
            import_text_code += "\n\nfrom wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D"
            import_text_code += "\n\nfrom wofryimpl.propagator.propagators1D.fresnel import Fresnel1D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators1D.fresnel_convolution import FresnelConvolution1D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators1D.fraunhofer import Fraunhofer1D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators1D.integral import Integral1D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators1D.fresnel_zoom import FresnelZoom1D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators1D.fresnel_zoom_scaling_theorem import FresnelZoomScaling1D"
        elif source_dimension == 2:
            import_text_code += "\n\nfrom wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D"
            import_text_code += "\n\nfrom wofryimpl.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators2D.fresnel import Fresnel2D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators2D.fresnel_convolution import FresnelConvolution2D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators2D.fraunhofer import Fraunhofer2D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators2D.integral import Integral2D"
            import_text_code +=   "\nfrom wofryimpl.propagator.propagators2D.fresnel_zoom_xy import FresnelZoomXY2D"

        source_text_code = ""
        if do_plot:
            source_text_code += "\n\nfrom srxraylib.plot.gol import plot, plot_image"
            source_text_code += "\nplot_from_oe = 0 # set to a large number to avoid plots"


        source_text_code  +=  "\n\n\n##########  SOURCE ##########\n\n\n"
        source_text_code += self.get_light_source().to_python_code(do_plot=False, add_import_section=False)


        if do_plot:
            if source_dimension == 1:
                source_text_code += "\n\n\nif plot_from_oe <= 0: plot(output_wavefront.get_abscissas(),output_wavefront.get_intensity(),title='SOURCE')"
            elif source_dimension == 2:
                source_text_code += "\n\n\nif plot_from_oe <= 0: plot_image(output_wavefront.get_intensity(),output_wavefront.get_coordinate_x(),output_wavefront.get_coordinate_y(),aspect='auto',title='SOURCE')"


    beamline_text_code = ""
    if self.get_beamline_elements_number() > 0:
        beamline_text_code += "\n\n\n##########  OPTICAL SYSTEM ##########\n\n\n"


        for index in range(self.get_beamline_elements_number()):
            beamline_text_code += "\n\n\n##########  OPTICAL ELEMENT NUMBER %i ##########\n\n\n" % (index+1)
            oe_name = "oe_" + str(index)
            beamline_element = self.get_beamline_element_at(index)
            optical_element = beamline_element.get_optical_element()
            coordinates = beamline_element.get_coordinates()

            beamline_text_code += "\ninput_wavefront = output_wavefront.duplicate()"

            # OPTICAL ELEMENT ----------------
            beamline_text_code += optical_element.to_python_code()

            propagation_info = self.get_propagation_info_at(index)

            if (coordinates.p() == 0.0) and (coordinates.q() == 0.0): # NO DRIFT
                beamline_text_code += "\n# no drift in this element "
                beamline_text_code += "\noutput_wavefront = optical_element.applyOpticalElement(input_wavefront)"
            else:
                if coordinates.p() != 0.0:
                    beamline_text_code += "\n# drift_before %g m" % coordinates.p()
                if coordinates.q() != 0.0:
                    beamline_text_code += "\n# drift_after %g m " % coordinates.q()

                ##########################
                # 1D
                # ==
                #
                # propagators_list = ["Fresnel",    "Fresnel (Convolution)",  "Fraunhofer",    "Integral",    "Fresnel Zoom",    "Fresnel Zoom Scaled"]
                # class_name       = ["Fresnel1D",  "FresnelConvolution1D",   "Fraunhofer1D",  "Integral1D",  "FresnelZoom1D",   "FresnelZoomScaling1D"]
                # handler_name     = ["FRESNEL_1D", "FRESNEL_CONVOLUTION_1D", "FRAUNHOFER_1D", "INTEGRAL_1D", "FRESNEL_ZOOM_1D", "FRESNEL_ZOOM_SCALING_1D"]
                #
                # 2D
                # ==
                # propagators_list = ["Fresnel",   "Fresnel (Convolution)",  "Fraunhofer",    "Integral",    "Fresnel Zoom XY"   ]
                # class_name       = ["Fresnel2D", "FresnelConvolution2D",   "Fraunhofer2D",  "Integral2D",  "FresnelZoomXY2D"   ]
                # handler_name     = ["FRESNEL_2D","FRESNEL_CONVOLUTION_2D", "FRAUNHOFER_2D", "INTEGRAL_2D", "FRESNEL_ZOOM_XY_2D"]

                propagator_class_name                   = propagation_info["propagator_class_name"]
                propagator_handler_name                 = propagation_info["propagator_handler_name"]
                propagator_additional_parameters_names  = propagation_info["propagator_additional_parameters_names"]
                propagator_additional_parameters_values = propagation_info["propagator_additional_parameters_values"]

                beamline_text_code += "\n#"
                beamline_text_code += "\n# propagating\n#"
                beamline_text_code += "\n#"
                beamline_text_code += "\npropagation_elements = PropagationElements()"
                beamline_text_code += "\nbeamline_element = BeamlineElement(optical_element=optical_element,"
                beamline_text_code += "    coordinates=ElementCoordinates(p=%f," % (coordinates.p())
                beamline_text_code += "    q=%f," % (coordinates.q())
                beamline_text_code += "    angle_radial=numpy.radians(%f)," % (coordinates.angle_radial())
                beamline_text_code += "    angle_azimuthal=numpy.radians(%f)))" % (coordinates.angle_azimuthal())
                beamline_text_code += "\npropagation_elements.add_beamline_element(beamline_element)"
                beamline_text_code += "\npropagation_parameters = PropagationParameters(wavefront=input_wavefront,"
                beamline_text_code += "    propagation_elements = propagation_elements)"
                beamline_text_code += "\n#self.set_additional_parameters(propagation_parameters)"
                beamline_text_code += "\n#"

                for i in range(len(propagator_additional_parameters_names)):
                    beamline_text_code += "\npropagation_parameters.set_additional_parameters('%s', %s)" % \
                    (propagator_additional_parameters_names[i], str(propagator_additional_parameters_values[i]))

                beamline_text_code += "\n#"
                beamline_text_code += "\npropagator = PropagationManager.Instance()"
                beamline_text_code += "\ntry:"
                beamline_text_code += "\n    propagator.add_propagator(%s())" % propagator_class_name
                beamline_text_code += "\nexcept:"
                beamline_text_code += "\n    pass"
                beamline_text_code += "\noutput_wavefront = propagator.do_propagation(propagation_parameters=propagation_parameters,"
                beamline_text_code += "    handler_name='%s')" % (propagator_handler_name)

            if do_plot:
                beamline_text_code += "\n\n\n#\n#---- plots -----\n#"
                if source_dimension == 1:
                    beamline_text_code += "\nif plot_from_oe <= %d: plot(output_wavefront.get_abscissas(),output_wavefront.get_intensity(),title='OPTICAL ELEMENT NR %d')" % (index+1, index+1)
                elif source_dimension == 2:
                    beamline_text_code += "\nif plot_from_oe <= %d: plot_image(output_wavefront.get_intensity(),output_wavefront.get_coordinate_x(),output_wavefront.get_coordinate_y(),aspect='auto',title='OPTICAL ELEMENT NR %d')" % (index+1, index+1)

    full_text_code = import_text_code

    indent = '    '

    full_text_code += "\n\n\n#\n# SOURCE========================\n#\n\n"
    full_text_code += "\n\n\ndef run_source(my_mode_index=0):"

    if isinstance(self.get_light_source(), WOLightSourceCMD):
        full_text_code += "\n" + indent + "global coherent_mode_decomposition"
        full_text_code += "\n" + indent + "try:"
        full_text_code += "\n" + indent*2 + "if my_mode_index == 0: raise Exception()"
        full_text_code += "\n" + indent*2 + "tmp = coherent_mode_decomposition"
        full_text_code += "\n" + indent + "except:"

        indented = indent*2 + source_text_code.replace('\n', '\n' + indent*2)
        full_text_code += indented

        full_text_code += "\n" + indent + "output_wavefront = coherent_mode_decomposition.get_eigenvector_wavefront(my_mode_index)"
        full_text_code += "\n" + indent + "return output_wavefront"
    elif isinstance(self.get_light_source(), WOLightSource):
        indented = indent + source_text_code.replace('\n', '\n' + indent)
        full_text_code += indented
        full_text_code += "\n" + indent + "# previous command is useless but..."
        full_text_code += "\n" + indent + "output_wavefront.set_gaussian_hermite_mode(sigma_x=%g,amplitude=%g,mode_x=my_mode_index,shift=%g,beta=%g)" % \
               (self.get_light_source()._sigma_h,
                self.get_light_source()._amplitude,
                self.get_light_source()._gaussian_shift,
                self.get_light_source()._beta_h)

        full_text_code += "\n" + indent + "return output_wavefront"


    full_text_code += "\n\n\n#\n# BEAMLINE========================\n#\n\n"
    full_text_code += "\n\n\ndef run_beamline(output_wavefront):"
    indented = indent + beamline_text_code.replace('\n', '\n' + indent)
    full_text_code += indented
    full_text_code += "\n" + indent + "return output_wavefront"

    full_text_code += "\n\n\n#\n# MAIN FUNCTION========================\n#\n\n"
    full_text_code += "\n\n\ndef main():"

    full_text_code += "\n" + indent + "from srxraylib.plot.gol import plot, plot_image"
    full_text_code += "\n" + indent + "from orangecontrib.esrf.wofry.util.tally import TallyCoherentModes"
    full_text_code += "\n" + indent + ""
    full_text_code += "\n" + indent + "tally = TallyCoherentModes()"
    full_text_code += "\n" + indent + "for my_mode_index in range(%g):" % mode_index_max
    full_text_code += "\n" + indent * 2 + "output_wavefront = run_source(my_mode_index=my_mode_index)"
    full_text_code += "\n" + indent * 2 + "output_wavefront = run_beamline(output_wavefront)"
    full_text_code += "\n" + indent * 2 + "tally.append(output_wavefront)"
    full_text_code += "\n" + indent + ""


    if graph_file_flag == 0:
        dump_file = ""
    elif graph_file_flag == 1:
        dump_file = "%s_cross_spectral_density.png" % root_file_name
    elif graph_file_flag == 2:
        dump_file = "%s_cross_spectral_density.pdf" % root_file_name

    full_text_code += "\n\n" + indent + 'tally.plot_cross_spectral_density(show=%d,filename="%s")' % (show_graph_flag, dump_file)

    if graph_file_flag == 0:
        dump_file = ""
    elif graph_file_flag == 1:
        dump_file = "%s_spectral_density.png" % root_file_name
    elif graph_file_flag == 2:
        dump_file = "%s_spectral_density.pdf" % root_file_name
    full_text_code += "\n" + indent + 'tally.plot_spectral_density(show=%d,filename="%s")' % (show_graph_flag, dump_file)

    if graph_file_flag == 0:
        dump_file = ""
    elif graph_file_flag == 1:
        dump_file = "%s_occupation.png" % root_file_name
    elif graph_file_flag == 2:
        dump_file = "%s_occupation.pdf" % root_file_name
    full_text_code += "\n" + indent + 'tally.plot_occupation(show=%d,filename="%s")' % (show_graph_flag, dump_file)

    if data_file_flag:
        full_text_code += "\n\n" + indent + 'tally.save_spectral_density(filename="%s_spectral_density.dat")' % (root_file_name)
        full_text_code += "\n" + indent + 'tally.save_occupation(filename="%s_occupation.dat")' % (root_file_name)

    full_text_code += "\n\n\n#\n# MAIN========================\n#\n\n"
    full_text_code += "\n\n\nmain()"
    return full_text_code



if __name__ == "__main__":
    import sys

    a = QApplication(sys.argv)
    ow = DiagonalizePythonScript()
    ow.show()
    a.exec_()
    ow.saveSettings()